"""
"""

import json
import os
import warnings
from copy import deepcopy
from pathlib import Path
from typing import Dict, List, Optional, Sequence, Tuple, Union, Any

import numpy as np
from scipy import signal as SS
from scipy.io import loadmat, savemat
from torch.utils.data.dataset import Dataset
from tqdm.auto import tqdm

from ...._preprocessors import PreprocManager
from ....databases import MITDB as DR
from ....utils.misc import (
    ReprMixin,
    get_record_list_recursive3,
    list_sum,
    add_docstring,
)
from ....utils.utils_data import (
    ensure_siglen,
    cls_to_bin,
    generate_weight_mask,
    mask_to_intervals,
)
from ....utils.utils_signal import remove_spikes_naive
from ....cfg import CFG, DEFAULTS
from .mitdb_cfg import MITDBTrainCfg


__all__ = [
    "MITDBDataset",
]


class MITDBDataset(ReprMixin, Dataset):
    """Data generator for feeding data into pytorch models
    using the :class:`~torch_ecg.databases.MITDB` database.

    Parameters
    ----------
    config : dict
        Configurations for the dataset, ref. `MITDBDataset`.
        A simple example is as follows:

        .. code-block:: python

            >>> config = deepcopy(MITDBDataset)
            >>> config.db_dir = "some/path/to/db"
            >>> dataset = MITDBDataset(config, task="qrs_detection", training=True, lazy=False)

    training : bool, default True
        If True, the training set will be loaded,
        otherwise the test (val) set will be loaded.
    lazy : bool, default True
        If True, the data will not be loaded immediately,
        instead, it will be loaded on demand.
    **reader_kwargs : dict, optional
        Keyword arguments for the database reader class.

    """

    __name__ = "MITDBDataset"

    def __init__(
        self,
        config: CFG,
        task: str,
        training: bool = True,
        lazy: bool = True,
        **reader_kwargs: Any,
    ) -> None:
        super().__init__()
        self.config = deepcopy(config)
        if reader_kwargs.pop("db_dir", None) is not None:
            warnings.warn(
                "`db_dir` is specified in both config and reader_kwargs", RuntimeWarning
            )
        self.reader = DR(db_dir=self.config.db_dir, **reader_kwargs)
        # assert self.config.db_dir is not None, "db_dir must be specified"
        self.config.db_dir = self.reader.db_dir
        self.dtype = self.config.np_dtype
        self.allowed_preproc = list(
            set(
                [
                    "bandpass",
                    "baseline_remove",
                    "normalize",
                ]
            ).intersection(set(self.config.keys()))
        )

        self.training = training

        self.lazy = lazy

        ppm_config = CFG(random=False)
        ppm_config.update(deepcopy(self.config))
        # ppm_config.pop("normalize")
        seg_ppm_config = CFG(random=False)
        seg_ppm_config.update(deepcopy(self.config))
        seg_ppm_config.pop("bandpass", None)
        self.ppm = PreprocManager.from_config(ppm_config)
        # self.ppm.rearrange(["bandpass", "baseline_remove", "normalize"])
        self.seg_ppm = PreprocManager.from_config(seg_ppm_config)

        # create directories if needed
        # segments_dir for sliced segments of fixed length
        self.segments_base_dir = self.config.db_dir / "segments"
        self.segments_base_dir.mkdir(parents=True, exist_ok=True)
        self.segment_name_pattern = "S_[\\d]{3}_[\\d]{7}"
        self.segment_ext = "mat"
        # rr_dir for sequence of rr intervals of fix length
        self.rr_seq_base_dir = self.config.db_dir / "rr_seq"
        self.rr_seq_base_dir.mkdir(parents=True, exist_ok=True)
        self.rr_seq_name_pattern = "R_[\\d]{3}_[\\d]{7}"
        self.rr_seq_ext = "mat"

        self._all_data = None
        self._all_labels = None
        self._all_masks = None
        self.__set_task(task, lazy=self.lazy)

    def _load_all_data(self) -> None:
        """Load all data into memory."""
        self.__set_task(self.task, lazy=False)

    def __set_task(self, task: str, lazy: bool = True) -> None:
        """Set the task and load the data if needed.

        Parameters
        ----------
        task : str
            Name of the task, as listed in `MITDBTrainCfg.tasks`.
        lazy : bool, default True
            If True, the data will not be loaded immediately,
            instead, it will be loaded on demand.

        Returns
        -------
        None

        """
        assert task.lower() in MITDBTrainCfg.tasks, f"illegal task \042{task}\042"
        if (
            hasattr(self, "task")
            and self.task == task.lower()
            and self._all_data is not None
            and len(self._all_data) > 0
        ):
            return
        self.task = task.lower()
        self.all_classes = self.config[task].classes
        self.n_classes = len(self.config[task].classes)
        self.lazy = lazy

        self.seglen = self.config[task].input_len  # alias, for simplicity
        split_res = self._train_test_split(self.task)
        if self.training:
            self.records = split_res.train
        else:
            self.records = split_res.test

        if self.task in ["beat_classification"]:
            self._all_data, self._all_labels = [], []
            if self.lazy:
                return
            with tqdm(
                range(len(self.records)),
                desc="Loading data",
                unit="record",
                dynamic_ncols=True,
                mininterval=1.0,
            ) as pbar:
                for idx in pbar:
                    data = self.reader.load_data(self.records[idx])
                    data, _ = self.ppm(data, self.config.fs)
                    for beat in self.reader.load_beat_ann(self.records[idx]):
                        if beat.symbol not in self.config[self.task].classes:
                            continue
                        beat_data = ensure_siglen(
                            data[
                                :,
                                max(0, beat.index - self.config[self.task].winL) : min(
                                    data.shape[-1],
                                    beat.index + self.config[self.task].winR,
                                ),
                            ],
                            self.config[self.task].input_len,
                        )
                        self._all_data.append(beat_data)
                        self._all_labels.append(
                            self.config[self.task].class_map[beat.symbol]
                        )
                self._all_data = np.array(self._all_data)
                self._all_labels = np.array(self._all_labels)
                if self.config[self.task].loss not in ["CrossEntropyLoss"]:
                    self._all_labels = cls_to_bin(
                        self._all_labels, len(self.config[self.task].classes)
                    )
        elif self.task in [
            "qrs_detection",
            "rhythm_segmentation",
            "af_event",
        ]:
            # for qrs detection
            self.segments_dirs = CFG()
            self.__all_segments = CFG()
            self.segments_json = self.segments_base_dir / "segments.json"
            self._ls_segments()
            self.segments = list_sum([self.__all_segments[rec] for rec in self.records])
            # if self.__DEBUG__:
            #     self.segments = DEFAULTS.RNG_sample(
            #         self.segments, int(len(self.segments) * 0.01)
            #     ).tolist()
            if self.training:
                DEFAULTS.RNG.shuffle(self.segments)
            # preload data
            self.fdr = _FastDataReader(
                self.config,
                self.task,
                self.seg_ppm,
                self.segments_dirs,
                self.segments,
                self.segment_ext,
                self.reader.rhythm_types_map,
            )
            if self.lazy:
                return
            self._all_data, self._all_labels, self._all_masks = [], [], []
            if len(self.fdr) == 0:
                warnings.warn(
                    f"No data found for task {self.task}, slice the data first.",
                    RuntimeWarning,
                )
            with tqdm(
                range(len(self.fdr)),
                desc="Loading data",
                unit="record",
                dynamic_ncols=True,
                mininterval=1.0,
            ) as pbar:
                for idx in pbar:
                    d, l, m = self.fdr[idx]
                    self._all_data.append(d)
                    self._all_labels.append(l)
                    self._all_masks.append(m)
            self._all_data = np.array(self._all_data).astype(self.dtype)
            self._all_labels = np.array(self._all_labels).astype(self.dtype)
            if self.task == "qrs_detection":
                self._all_masks = None
            else:
                self._all_masks = np.array(self._all_masks).astype(self.dtype)
        elif self.task in [
            "rr_lstm",
        ]:
            self.rr_seq_dirs = CFG()
            self.__all_rr_seq = CFG()
            self.rr_seq_json = self.rr_seq_base_dir / "rr_seq.json"
            self._ls_rr_seq()
            self.rr_seq = list_sum([self.__all_rr_seq[rec] for rec in self.records])
            # if self.__DEBUG__:
            #     self.rr_seq = DEFAULTS.RNG_sample(
            #         self.rr_seq, int(len(self.rr_seq) * 0.01)
            #     ).tolist()
            if self.training:
                DEFAULTS.RNG.shuffle(self.rr_seq)
            # preload data
            self.fdr = _FastDataReader(
                self.config,
                self.task,
                self.seg_ppm,
                self.rr_seq_dirs,
                self.rr_seq,
                self.rr_seq_ext,
                self.reader.rhythm_types_map,
            )
            if self.lazy:
                return
            self._all_data, self._all_labels, self._all_masks = [], [], []
            with tqdm(
                range(len(self.fdr)),
                desc="Loading data",
                unit="record",
                dynamic_ncols=True,
                mininterval=1.0,
            ) as pbar:
                for idx in pbar:
                    d, l, m = self.fdr[idx]
                    self._all_data.append(d)
                    self._all_labels.append(l)
                    self._all_masks.append(m)
            self._all_data = np.array(self._all_data).astype(self.dtype)
            self._all_labels = np.array(self._all_labels).astype(self.dtype)
            self._all_masks = np.array(self._all_masks).astype(self.dtype)
        else:
            raise NotImplementedError(
                f"data generator for task \042{self.task}\042 not implemented"
            )

    def reset_task(self, task: str, lazy: bool = True) -> None:
        """Reset the task of the data generator.

        Parameters
        ----------
        task : str
            The task to be set.
        lazy : bool, optional
            Whether to load the data lazily, by default True.

        Returns
        -------
        None

        """
        self.__set_task(task, lazy)

    def _ls_segments(self) -> None:
        """Find all the segments in the segments base directory,
        and store them in some private attributes.
        """
        for item in ["data", "ann"]:
            self.segments_dirs[item] = CFG()
            for rec in self.reader:
                self.segments_dirs[item][rec] = self.segments_base_dir / item / rec
                self.segments_dirs[item][rec].mkdir(parents=True, exist_ok=True)
        if self.segments_json.is_file():
            self.__all_segments = json.loads(self.segments_json.read_text())
            # return
        print(
            "please allow the reader a few minutes to collect "
            f"the segments from {self.segments_base_dir}..."
        )
        seg_filename_pattern = f"{self.segment_name_pattern}\\.{self.segment_ext}"
        self.__all_segments = CFG(
            {
                rec: get_record_list_recursive3(
                    str(self.segments_dirs.data[rec]), seg_filename_pattern
                )
                for rec in self.reader
            }
        )
        if all([len(self.__all_segments[rec]) > 0 for rec in self.reader]):
            self.segments_json.write_text(
                json.dumps(self.__all_segments, ensure_ascii=False)
            )

    def _ls_rr_seq(self) -> None:
        """Find all the rr sequences in the rr sequences base directory,
        and store them in some private attributes.
        """
        for rec in self.reader:
            self.rr_seq_dirs[rec] = self.rr_seq_base_dir / rec
            self.rr_seq_dirs[rec].mkdir(parents=True, exist_ok=True)
        if self.rr_seq_json.is_file():
            self.__all_rr_seq = json.loads(self.rr_seq_json.read_text())
            # return
        print(
            "please allow the reader a few minutes to collect "
            f"the rr sequences from {self.rr_seq_base_dir}..."
        )
        rr_seq_filename_pattern = f"{self.rr_seq_name_pattern}\\.{self.rr_seq_ext}"
        self.__all_rr_seq = CFG(
            {
                rec: get_record_list_recursive3(
                    self.rr_seq_dirs[rec], rr_seq_filename_pattern
                )
                for rec in self.reader
            }
        )
        if all([len(self.__all_rr_seq[rec]) > 0 for rec in self.reader]):
            self.rr_seq_json.write_text(
                json.dumps(self.__all_rr_seq, ensure_ascii=False)
            )

    @property
    def all_segments(self) -> CFG:
        if self.task in [
            "qrs_detection",
            "rhythm_segmentation",
            "af_event",
        ]:
            return self.__all_segments
        else:
            return CFG()

    @property
    def all_rr_seq(self) -> CFG:
        if self.task.lower() in [
            "rr_lstm",
        ]:
            return self.__all_rr_seq
        else:
            return CFG()

    def __len__(self) -> int:
        if self.task in ["beat_classification"]:
            return len(self._all_data)
        return len(self.fdr)

    def __getitem__(self, index: int) -> Tuple[np.ndarray, ...]:
        if self.task in ["beat_classification"]:
            return self._all_data[index], self._all_labels[index]
        if self.lazy:
            if self.task in ["qrs_detection"]:
                return self.fdr[index][:2]
            else:
                return self.fdr[index]
        else:
            if self.task in ["qrs_detection"]:
                return self._all_data[index], self._all_labels[index]
            else:
                return (
                    self._all_data[index],
                    self._all_labels[index],
                    self._all_masks[index],
                )

    def _get_seg_data_path(self, seg: str) -> Path:
        """Get the path of the data file of the segment.

        Parameters
        ----------
        seg : str
            Name of the segment, of pattern like "S_100_0000193".

        Returns
        -------
        pathlib.Path
            Absolute path of the data file of the segment.

        """
        rec = self._get_rec_name(seg)
        fp = self.segments_dirs.data[rec] / f"{seg}.{self.segment_ext}"
        return fp

    @add_docstring(_get_seg_data_path.__doc__.replace("data file", "annotation file"))
    def _get_seg_ann_path(self, seg: str) -> Path:
        rec = self._get_rec_name(seg)
        fp = self.segments_dirs.ann[rec] / f"{seg}.{self.segment_ext}"
        return fp

    def _load_seg_data(self, seg: str) -> np.ndarray:
        """Load data of the segment.

        Parameters
        ----------
        seg : str
            Name of the segment, of pattern like "S_100_0000193".

        Returns
        -------
        numpy.ndarray
            Data of the segment, of shape ``(2, self.seglen)``.

        """
        seg_data_fp = self._get_seg_data_path(seg)
        seg_data = loadmat(str(seg_data_fp))["ecg"]
        return seg_data

    def _load_seg_ann(self, seg: str) -> dict:
        """Load annotations of the segment.

        Parameters
        ----------
        seg : str
            Name of the segment, of pattern like "S_100_0000193".

        Returns
        -------
        dict
            A dictionay of annotations of the segment, including

                - rpeaks: indices of rpeaks of the segment
                - qrs_mask: mask of qrs complexes of the segment
                - rhythm_mask: mask of rhythms of the segment
                - interval: interval ([start_idx, end_idx]) in the
                  original ECG record of the segment

        """
        seg_ann_fp = self._get_seg_ann_path(seg)
        seg_ann = {
            k: v.flatten()
            for k, v in loadmat(str(seg_ann_fp)).items()
            if not k.startswith("__")
        }
        return seg_ann

    def _load_seg_mask(
        self, seg: str, task: Optional[str] = None
    ) -> Union[np.ndarray, Dict[str, np.ndarray]]:
        """Load mask(s) of the segment.

        Parameters
        ----------
        seg : str
            Name of the segment, of pattern like "S_100_0000193".
        task : str, optional
            Task for doing the segmentation, by default the current task.
            If specified, overrides self.task.
            If is "all", then all masks
            ("qrs_mask", "rhythm_mask", etc.) will be returned.

        Returns
        -------
        numpy.ndarray or dict
            Mask(s) of the segment,
            of shape ``(self.seglen, self.n_classes)``.

        """
        seg_mask = {
            k: v.reshape((self.seglen, -1))
            for k, v in self._load_seg_ann(seg).items()
            if k
            in [
                "qrs_mask",
                "rhythm_mask",
            ]
        }
        _task = (task or self.task).lower()
        if _task == "all":
            return seg_mask
        if _task in [
            "qrs_detection",
        ]:
            seg_mask = seg_mask["qrs_mask"]
        elif _task in ["rhythm_segmentation", "af_event"]:
            seg_mask = seg_mask["rhythm_mask"]
        return seg_mask

    def _load_seg_seq_lab(self, seg: str, reduction: int) -> np.ndarray:
        """Load sequence label of the segment.

        Parameters
        ----------
        seg : str
            Name of the segment, of pattern like "S_100_0000193".
        reduction : int
            Reduction ratio (granularity) of length of the model output,
            compared to the original signal length.

        Returns
        -------
        numpy.ndarray
            Label of the sequence,
            of shape ``(self.seglen//reduction, self.n_classes)``.

        """
        seg_mask = self._load_seg_mask(seg)
        seg_len, n_classes = seg_mask.shape
        seq_lab = np.stack(
            arrays=[
                np.mean(
                    seg_mask[reduction * idx : reduction * (idx + 1)],
                    axis=0,
                    keepdims=True,
                ).astype(int)
                for idx in range(seg_len // reduction)
            ],
            axis=0,
        ).squeeze(axis=1)
        return seq_lab

    def _get_rr_seq_path(self, seq_name: str) -> Path:
        """Get path of the data file of the rr_seq.

        Parameters
        ----------
        seq_name : str
            Name of the rr_seq, of pattern like "R_100_0000193".

        Returns
        -------
        pathlib.Path
            Absolute path of the data file of the rr_seq.

        """
        rec = self._get_rec_name(seq_name)
        fp = self.rr_seq_dirs[rec] / f"{seq_name}.{self.rr_seq_ext}"
        return fp

    def _load_rr_seq(self, seq_name: str) -> Dict[str, np.ndarray]:
        """Load metadata of sequence of rr intervals.

        Parameters
        ----------
        seq_name : str
            Name of the rr_seq, of pattern like "R_100_0000193".

        Returns
        -------
        dict
            Metadata of sequence of rr intervals, including

                - rr: the sequence of rr intervals, with units in seconds,
                  of shape ``(self.seglen, 1)``
                - label: label of the rr intervals,
                  of shape ``(self.seglen, self.n_classes)``
                - interval: interval of the current rr sequence
                  in the whole rr sequence in the original record

        """
        rr_seq_path = self._get_rr_seq_path(seq_name)
        rr_seq = {
            k: v for k, v in loadmat(str(rr_seq_path)).items() if not k.startswith("__")
        }
        rr_seq["rr"] = rr_seq["rr"].reshape((self.seglen, 1))
        rr_seq["label"] = rr_seq["label"].reshape((self.seglen, self.n_classes))
        rr_seq["interval"] = rr_seq["interval"].flatten()
        return rr_seq

    def persistence(self, force_recompute: bool = False, verbose: int = 0) -> None:
        """Save the preprocessed data to disk.

        Parameters
        ----------
        force_recompute : bool, default False
            Whether to force recompute the preprocessed data.
        verbose : int, default 0
            Verbosity level for printing the progress.

        Returns
        -------
        None

        """
        # TODO: consider whether preprocessing should be added
        # if verbose >= 1:
        #     print(" preprocessing data ".center(110, "#"))
        # self._preprocess_data(
        #     force_recompute=force_recompute,
        #     verbose=verbose,
        # )

        original_task = self.task
        original_lazy = self.lazy
        self.__set_task("qrs_detection", lazy=True)
        if verbose >= 1:
            print("\n" + " slicing data into segments ".center(110, "#"))
        self._slice_data(
            force_recompute=force_recompute,
            verbose=verbose,
        )

        self.__set_task("rr_lstm", lazy=True)
        if verbose >= 1:
            print("\n" + " generating rr sequences ".center(110, "#"))
        self._slice_rr_seq(
            force_recompute=force_recompute,
            verbose=verbose,
        )

        self.__set_task(original_task, lazy=original_lazy)

    def _slice_data(self, force_recompute: bool = False, verbose: int = 0) -> None:
        """Slice all records into segments.

        Slice all records into segments of length `self.seglen`,
        and perform data augmentations specified in `self.config`.

        Parameters
        ----------
        force_recompute : bool, default False
            Whether to force recompute the preprocessed data.
        verbose : int, default 0
            Verbosity level for printing the progress.

        Returns
        -------
        None

        """
        self.__assert_task(
            [
                "qrs_detection",
                "rhythm_segmentation",
                "af_event",  # segmentation of AF events
            ]
        )
        if force_recompute:
            self._clear_cached_segments()
        with tqdm(
            enumerate(self.reader),
            total=len(self.reader),
            desc="Slicing data",
            unit="record",
            dynamic_ncols=True,
            mininterval=1.0,
        ) as pbar:
            for idx, rec in pbar:
                self._slice_one_record(
                    rec=rec,
                    force_recompute=False,
                    update_segments_json=False,
                    verbose=verbose,
                )
                # if verbose >= 1:
                #     print(f"{idx+1}/{len(self.reader)} records", end="\r")
        if force_recompute:
            self.segments_json.write_text(
                json.dump(self.__all_segments, ensure_ascii=False)
            )

    def _slice_one_record(
        self,
        rec: str,
        force_recompute: bool = False,
        update_segments_json: bool = False,
        verbose: int = 0,
    ) -> None:
        """Slice one record into segments.

        Slice one record into segments of length `self.seglen`,
        and perform data augmentations specified in `self.config`.

        Parameters
        ----------
        rec : str
            Name of the record.
        force_recompute : bool, default False
            Whether to force recompute the preprocessed data.
        update_segments_json : bool, default False
            If both `force_recompute` and `update_segments_json` are True,
            the file `self.segments_json` will be updated.
            Useful when slicing not all records.
        verbose : int, default 0
            Verbosity level for printing the progress.

        Returns
        -------
        None

        """
        self.__assert_task(
            [
                "qrs_detection",
                "rhythm_segmentation",
                "af_event",  # segmentation of AF events
            ]
        )
        rec_segs = self.__all_segments[rec]
        if (not force_recompute) and len(rec_segs) > 0:
            return
        elif force_recompute:
            self._clear_cached_segments([rec])

        data, _ = self.ppm(self.reader.load_data(rec), self.config.fs)
        siglen = data.shape[1]
        rpeaks = self.reader.load_rpeak_indices(rec)
        rhythm_mask = self.reader.load_rhythm_ann(
            rec,
            rhythm_format="mask",
            rhythm_types=self.config.rhythm_segmentation.classes,
        )
        forward_len = self.seglen - self.config[self.task].overlap_len
        critical_forward_len = self.seglen - self.config[self.task].critical_overlap_len
        critical_forward_len = [critical_forward_len // 4, critical_forward_len]

        # find critical points
        critical_points = np.where(np.diff(rhythm_mask) != 0)[0]
        critical_points = [
            p
            for p in critical_points
            if critical_forward_len[1] <= p < siglen - critical_forward_len[1]
        ]

        segments = []

        # ordinary segments with constant forward_len
        print("Slicing ordinary segments with constant `forward_len`")
        with tqdm(
            range((siglen - self.seglen) // forward_len + 1),
            desc=f"Slicing segments for record {rec}",
            unit="segment",
            dynamic_ncols=True,
            mininterval=1.0,
        ) as pbar:
            for idx in pbar:
                start_idx = idx * forward_len
                new_seg = self.__generate_segment(
                    rec=rec,
                    data=data,
                    start_idx=start_idx,
                )
                segments.append(new_seg)
        # the tail segment
        new_seg = self.__generate_segment(
            rec=rec,
            data=data,
            end_idx=siglen,
        )
        segments.append(new_seg)

        if len(critical_points) == 0:
            # save segments
            self.__save_segments(rec, segments, update_segments_json)
            return

        # special segments around critical_points with random forward_len in critical_forward_len
        print(
            "Slicing special segments around `critical_points` "
            "with random `forward_len` in `critical_forward_len`"
        )
        with tqdm(
            critical_points,
            desc=f"Slicing segments for record {rec}",
            unit="segment",
            dynamic_ncols=True,
            mininterval=1.0,
        ) as pbar:
            for cp in pbar:
                start_idx = max(
                    0,
                    cp
                    - self.seglen
                    + DEFAULTS.RNG_randint(
                        critical_forward_len[0], critical_forward_len[1]
                    ),
                )
                while start_idx <= min(
                    cp - critical_forward_len[1], siglen - self.seglen
                ):
                    new_seg = self.__generate_segment(
                        rec=rec,
                        data=data,
                        start_idx=start_idx,
                    )
                    segments.append(new_seg)
                    start_idx += DEFAULTS.RNG_randint(
                        critical_forward_len[0], critical_forward_len[1]
                    )

        # save segments
        self.__save_segments(rec, segments, update_segments_json)

    def __generate_segment(
        self,
        rec: str,
        data: np.ndarray,
        start_idx: Optional[int] = None,
        end_idx: Optional[int] = None,
    ) -> CFG:
        """Generate segment, with possible data augmentation.

        Parameter
        ---------
        rec : str
            Name of the record.
        data : numpy.ndarray
            The whole of (preprocessed) ECG record.
        start_idx : int, optional
            Start index of the signal for generating the segment.
        end_idx : int, optional
            End index of the signal for generating the segment.
            If `start_idx` is set, then `end_idx` is ignored,
            since the segment length is fixed to `self.seglen`.
            At least one of `start_idx` and `end_idx` should be set.

        Returns
        -------
        dict
            Segments (meta-)data, containing:

                - data: values of the segment, with units in mV
                - rpeaks: indices of rpeaks of the segment
                - qrs_mask: mask of qrs complexes of the segment
                - rhythm_mask: mask of rhythms of the segment
                - interval: interval ([start_idx, end_idx]) in the
                  original ECG record of the segment

        """
        assert not all(
            [start_idx is None, end_idx is None]
        ), "at least one of `start_idx` and `end_idx` should be set"
        siglen = data.shape[1]
        # offline augmentations are done, including strech-or-compress, ...
        if self.config.stretch_compress != 0:
            stretch_compress_choices = [0, 1, -1]
            sign = DEFAULTS.RNG_sample(stretch_compress_choices, 1)[0]
            if sign != 0:
                sc_ratio = self.config.stretch_compress
                sc_ratio = (
                    1 + (DEFAULTS.RNG.uniform(sc_ratio / 4, sc_ratio) * sign) / 100
                )
                sc_len = int(round(sc_ratio * self.seglen))
                if start_idx is not None:
                    end_idx = start_idx + sc_len
                else:
                    start_idx = end_idx - sc_len
                if end_idx > siglen:
                    end_idx = siglen
                    start_idx = max(0, end_idx - sc_len)
                    sc_ratio = (end_idx - start_idx) / self.seglen
                aug_seg = data[..., start_idx:end_idx]
                aug_seg = SS.resample(x=aug_seg, num=self.seglen, axis=1)
            else:
                if start_idx is not None:
                    end_idx = start_idx + self.seglen
                    if end_idx > siglen:
                        end_idx = siglen
                        start_idx = end_idx - self.seglen
                else:
                    start_idx = end_idx - self.seglen
                    if start_idx < 0:
                        start_idx = 0
                        end_idx = self.seglen
                # the segment of original signal, with no augmentation
                aug_seg = data[..., start_idx:end_idx]
                sc_ratio = 1
        else:
            if start_idx is not None:
                end_idx = start_idx + self.seglen
                if end_idx > siglen:
                    end_idx = siglen
                    start_idx = end_idx - self.seglen
            else:
                start_idx = end_idx - self.seglen
                if start_idx < 0:
                    start_idx = 0
                    end_idx = self.seglen
            aug_seg = data[..., start_idx:end_idx]
            sc_ratio = 1
        # adjust rpeaks
        seg_rpeaks = self.reader.load_rpeak_indices(
            rec=rec,
            sampfrom=start_idx,
            sampto=end_idx,
            keep_original=False,
        )
        seg_rpeaks = [
            int(round(r / sc_ratio))
            for r in seg_rpeaks
            if self.config.rpeaks_dist2border
            <= r
            < self.seglen - self.config.rpeaks_dist2border
        ]
        # generate qrs_mask from rpeaks
        seg_qrs_mask = np.zeros((self.seglen,), dtype=int)
        for r in seg_rpeaks:
            seg_qrs_mask[
                r - self.config.qrs_mask_bias : r + self.config.qrs_mask_bias
            ] = 1
        # adjust rhythm_intervals
        seg_rhythm_intervals = self.reader.load_rhythm_ann(
            rec=rec,
            sampfrom=start_idx,
            sampto=end_idx,
            rhythm_format="intervals",
            rhythm_types=self.config.rhythm_segmentation.classes,
            keep_original=False,
        )
        seg_rhythm_intervals = {
            rt: [
                [int(round(itv[0] / sc_ratio)), int(round(itv[1] / sc_ratio))]
                for itv in l_itvs
            ]
            for rt, l_itvs in seg_rhythm_intervals.items()
        }
        # generate rhythm_mask from rhythm_intervals
        seg_rhythm_mask = np.zeros((self.seglen,), dtype=int)
        for rt, l_itvs in seg_rhythm_intervals.items():
            for itv in l_itvs:
                seg_rhythm_mask[
                    itv[0] : itv[1]
                ] = self.config.rhythm_segmentation.class_map[rt]

        new_seg = CFG(
            data=aug_seg,
            rpeaks=seg_rpeaks,
            qrs_mask=seg_qrs_mask,
            rhythm_mask=seg_rhythm_mask,
            interval=[start_idx, end_idx],
        )
        return new_seg

    def __save_segments(
        self, rec: str, segments: List[CFG], update_segments_json: bool = False
    ) -> None:
        """Save the segments to disk.

        Parameters
        ----------
        rec : str
            Name of the record
        segments : List[dict]
            List of the segments (meta-)data to be saved.
        update_segments_json : bool, default False
            Whether to update the segments.json file.

        """
        ordering = list(range(len(segments)))
        DEFAULTS.RNG.shuffle(ordering)
        for i, idx in enumerate(ordering):
            seg = segments[idx]
            filename = f"S_{rec}_{i:07d}.{self.segment_ext}"
            data_path = self.segments_dirs.data[rec] / filename
            savemat(str(data_path), {"ecg": seg.data})
            self.__all_segments[rec].append(Path(filename).with_suffix("").name)
            ann_path = self.segments_dirs.ann[rec] / filename
            savemat(
                str(ann_path),
                {
                    k: v
                    for k, v in seg.items()
                    if k
                    not in [
                        "data",
                    ]
                },
            )
        if update_segments_json:
            self.segments_json.write_text(
                json.dumps(self.__all_segments, ensure_ascii=False)
            )

    def _clear_cached_segments(self, recs: Optional[Sequence[str]] = None) -> None:
        """Clear the cached segments of the records.

        Parameters
        ----------
        recs : Sequence[str], optional
            Sequence of the records whose segments are to be cleared.
            Defaults to all records.

        Returns
        -------
        None

        """
        self.__assert_task(
            [
                "qrs_detection",
                "rhythm_segmentation",
                "af_event",  # segmentation of AF events
            ]
        )
        if recs is None:
            recs = self.reader.all_records
        for rec in recs:
            for item in [
                "data",
                "ann",
            ]:
                path = str(self.segments_dirs[item][rec])
                for f in [n for n in os.listdir(path) if n.endswith(self.segment_ext)]:
                    os.remove(os.path.join(path, f))
                    if os.path.splitext(f)[0] in self.__all_segments[rec]:
                        self.__all_segments[rec].remove(os.path.splitext(f)[0])
        self.segments = list_sum([self.__all_segments[rec] for rec in self.records])

    def _slice_rr_seq(self, force_recompute: bool = False, verbose: int = 0) -> None:
        """Slice sequences of rr intervals into fixed length (sub)sequences.

        Parameters
        ----------
        force_recompute : bool, default False
            Whether to force recompute the rr sequences.
        verbose : int, default 0
            Verbosity level for printing the progress.

        Returns
        -------
        None

        """
        self.__assert_task(["rr_lstm"])
        if force_recompute:
            self._clear_cached_rr_seq()
        with tqdm(
            enumerate(self.reader),
            total=len(self.reader),
            desc="Slicing rr_seq",
            unit="record",
            dynamic_ncols=True,
            mininterval=1.0,
        ) as pbar:
            for idx, rec in pbar:
                self._slice_rr_seq_one_record(
                    rec=rec,
                    force_recompute=False,
                    update_rr_seq_json=False,
                    verbose=verbose,
                )
                # if verbose >= 1:
                #     print(f"{idx+1}/{len(self.reader.all_records)} records", end="\r")
        if force_recompute:
            self.rr_seq_json.write_text(
                json.dumps(self.__all_rr_seq, ensure_ascii=False)
            )

    def _slice_rr_seq_one_record(
        self,
        rec: str,
        force_recompute: bool = False,
        update_rr_seq_json: bool = False,
        verbose: int = 0,
    ) -> None:
        """Slice sequences of rr intervals into
        fixed length (sub)sequences for one record.

        Parameters
        ----------
        rec : str
            Name of the record
        force_recompute : bool, default False,
            If True, the rr sequences will be recomputed regardless of
            whether they have been computed before.
        update_rr_seq_json : bool, default False
            Whether to update the rr_seq.json file.
        verbose : int, default 0
            Verbosity level for printing the progress.

        Returns
        -------
        None

        """
        self.__assert_task(["rr_lstm"])
        rec_rr_seq = self.__all_rr_seq[rec]
        if (not force_recompute) and len(rec_rr_seq) > 0:
            return
        elif force_recompute:
            self._clear_cached_rr_seq([rec])

        forward_len = self.seglen - self.config[self.task].overlap_len
        critical_forward_len = self.seglen - self.config[self.task].critical_overlap_len
        critical_forward_len = [critical_forward_len - 2, critical_forward_len]

        rpeaks = self.reader.load_rpeak_indices(rec)
        rr = np.diff(rpeaks) / self.config.fs
        if len(rr) < self.seglen:
            return
        rhythm_mask = self.reader.load_rhythm_ann(
            rec,
            rhythm_format="mask",
            rhythm_types=self.config.rhythm_segmentation.classes,
        )
        label_seq = rhythm_mask[rpeaks][:-1]

        # find critical points
        critical_points = np.where(np.diff(label_seq) != 0)[0]
        critical_points = [
            p
            for p in critical_points
            if critical_forward_len[1] <= p < len(rr) - critical_forward_len[1]
        ]

        rr_seq = []

        # ordinary rr_seq with constant forward_len
        print("Slicing ordinary rr_seq with constant `forward_len`")
        with tqdm(
            range((len(rr) - self.seglen) // forward_len + 1),
            desc=f"Slicing rr_seq for record `{rec}`",
            unit="segment",
            dynamic_ncols=True,
            mininterval=1.0,
        ) as pbar:
            for idx in pbar:
                start_idx = idx * forward_len
                end_idx = start_idx + self.seglen
                new_rr_seq = CFG(
                    rr=rr[start_idx:end_idx],
                    label=label_seq[start_idx:end_idx],
                    interval=[start_idx, end_idx],
                )
                rr_seq.append(new_rr_seq)
        # the tail segment
        if end_idx < len(rr):
            end_idx = len(rr)
            start_idx = end_idx - self.seglen
            new_rr_seq = CFG(
                rr=rr[start_idx:end_idx],
                label=label_seq[start_idx:end_idx],
                interval=[start_idx, end_idx],
            )
            rr_seq.append(new_rr_seq)

        if len(critical_points) == 0:
            # save rr sequences
            self.__save_rr_seq(rec, rr_seq, update_rr_seq_json)
            return

        # special rr_seq around critical_points with random forward_len in critical_forward_len
        print(
            "Slicing special rr_seq around `critical_points` "
            "with random `forward_len` in `critical_forward_len`"
        )
        with tqdm(
            critical_points,
            desc=f"Slicing rr_seq for record {rec}",
            unit="segment",
            dynamic_ncols=True,
            mininterval=1.0,
        ) as pbar:
            for cp in pbar:
                start_idx = max(
                    0,
                    cp
                    - self.seglen
                    + DEFAULTS.RNG_randint(
                        critical_forward_len[0], critical_forward_len[1]
                    ),
                )
                while start_idx <= min(
                    cp - critical_forward_len[1], len(rr) - self.seglen
                ):
                    end_idx = start_idx + self.seglen
                    new_rr_seq = CFG(
                        rr=rr[start_idx:end_idx],
                        label=label_seq[start_idx:end_idx],
                        interval=[start_idx, end_idx],
                    )
                    rr_seq.append(new_rr_seq)
                    start_idx += DEFAULTS.RNG_randint(
                        critical_forward_len[0], critical_forward_len[1]
                    )
        # save rr sequences
        self.__save_rr_seq(rec, rr_seq, update_rr_seq_json)

    def __save_rr_seq(
        self, rec: str, rr_seq: List[CFG], update_rr_seq_json: bool = False
    ) -> None:
        """Save rr_seq to disk.

        Parameters
        ----------
        rec : str
            Name of the record.
        rr_seq : List[dict]
            List of the rr_seq (meta-)data to be saved.
        update_rr_seq_json : bool, default False
            Whether to update the rr_seq.json file.

        Returns
        -------
        None

        """
        ordering = list(range(len(rr_seq)))
        DEFAULTS.RNG.shuffle(ordering)
        for i, idx in enumerate(ordering):
            item = rr_seq[idx]
            filename = f"R_{rec}_{i:07d}.{self.rr_seq_ext}"
            data_path = self.rr_seq_dirs[rec] / filename
            savemat(str(data_path), item)
            self.__all_rr_seq[rec].append(Path(filename).with_suffix("").name)
        if update_rr_seq_json:
            self.rr_seq_json.write_text(
                json.dumps(self.__all_rr_seq, ensure_ascii=False)
            )

    def _clear_cached_rr_seq(self, recs: Optional[Sequence[str]] = None) -> None:
        """Clear the cached rr sequences.

        Parameters
        ----------
        recs : Sequence[str], optional
            Sequence of the records whose segments are to be cleared.
            Defaults to all records.

        Returns
        -------
        None

        """
        self.__assert_task(["rr_lstm"])
        if recs is None:
            recs = self.reader.all_records
        for rec in recs:
            path = str(self.rr_seq_dirs[rec])
            for f in [n for n in os.listdir(path) if n.endswith(self.rr_seq_ext)]:
                os.remove(os.path.join(path, f))
                if os.path.splitext(f)[0] in self.__all_rr_seq[rec]:
                    self.__all_rr_seq[rec].remove(os.path.splitext(f)[0])
        self.rr_seq = list_sum([self.__all_rr_seq[rec] for rec in self.records])

    def _get_rec_name(self, seg_or_rr: str) -> str:
        """Get the record name from the segment or rr_seq name.

        Parameters
        ----------
        seg_or_rr : str
            Name of the segment or rr_seq.

        Returns
        -------
        rec : str
            Name of the record
            that the segment or rr_seq was generated from.

        """
        rec = seg_or_rr.split("_")[1]
        return rec

    def _train_test_split(self, task: str) -> Dict[str, List[str]]:
        """Perform train-test split.

        Parameters
        ----------
        task : str
            Task name for which the split is performed.

        Returns
        -------
        split_res : dict
            A dictionary of the split results.
            Keys are "train" and "test", and
            values are list of the subjects split for training or validation.

        """
        if task in [
            "beat_classification",
            "qrs_detection",
        ]:
            full_test_set = ["101", "102", "108", "114", "207", "223"]
            train_set = [rec for rec in self.reader if rec not in full_test_set]
            test_set = [rec for rec in self.reader if rec in full_test_set]
        else:  # rhythm segmentation, af event, rr_lstm
            full_test_set = ["106", "114", "124", "202", "217", "232"]
            train_set = [rec for rec in self.reader if rec not in full_test_set]
            test_set = [rec for rec in self.reader if rec in full_test_set]

        split_res = CFG(
            {
                "train": train_set,
                "test": test_set,
            }
        )
        return split_res

    def __assert_task(self, tasks: List[str]) -> None:
        """Check if the current task is in the given list."""
        assert self.task in tasks, (
            f"DO NOT call this method when the current task is {self.task}. "
            "Switch task using `reset_task`"
        )

    def plot_seg(self, seg: str, ticks_granularity: int = 0) -> None:
        """Plot the segment.

        Parameters
        ----------
        seg : str
            Name of the segment, of pattern like "S_1_1_0000193".
        ticks_granularity : int, default 0
            Granularity to plot axis ticks, the higher the more ticks.
            0 (no ticks) --> 1 (major ticks) --> 2 (major + minor ticks)

        Returns
        -------
        None

        """
        seg_data = self._load_seg_data(seg)
        seg_ann = self._load_seg_ann(seg)
        seg_ann["rhythm_intervals"] = mask_to_intervals(seg_ann["rhythm_mask"], vals=1)
        rec_name = self._get_rec_name(seg)
        self.reader.plot(
            rec=rec_name,  # unnecessary indeed
            data=seg_data,
            ann=seg_ann,
            ticks_granularity=ticks_granularity,
        )

    def extra_repr_keys(self) -> List[str]:
        return [
            "training",
            "task",
            "reader",
        ]


class _FastDataReader(ReprMixin, Dataset):
    """Fast data reader.

    Parameters
    ----------
    reader : CR
        The reader to read the data.
    records : Sequence[str]
        The list of records to read.
    config : CFG
        The configuration.
    ppm : PreprocManager, optional
        The preprocessor manager.

    """

    def __init__(
        self,
        config: CFG,
        task: str,
        seg_ppm: PreprocManager,
        file_dirs: dict,
        files: List[str],
        file_ext: str,
        rhythm_types_map: dict,
    ) -> None:
        self.config = config
        self.task = task
        self.seg_ppm = seg_ppm
        self.file_dirs = file_dirs
        self.files = files
        self.file_ext = file_ext
        self.rhythm_types_map = rhythm_types_map

        self.seglen = self.config[self.task].input_len
        self.n_classes = len(self.config[task].classes)

        self._seg_keys = {
            "qrs_detection": "qrs_mask",
            "rhythm_segmentation": "rhythm_mask",
            "af_event": "rhythm_mask",  # segmentation of AF events
        }

    def __getitem__(self, index: int) -> Tuple[np.ndarray, ...]:
        if self.task in [
            "qrs_detection",
            "rhythm_segmentation",
            "af_event",  # segmentation of AF events
        ]:
            seg_name = self.files[index]
            rec = seg_name.split("_")[1]
            seg_data_fp = self.file_dirs.data[rec] / f"{seg_name}.{self.file_ext}"
            seg_data = loadmat(str(seg_data_fp))["ecg"]
            for idx in range(seg_data.shape[0]):
                seg_data[idx] = remove_spikes_naive(seg_data[idx])
            seg_ann_fp = self.file_dirs.ann[rec] / f"{seg_name}.{self.file_ext}"
            seg_label = loadmat(str(seg_ann_fp))[self._seg_keys[self.task]].reshape(
                (self.seglen, -1)
            )
            if self.config[self.task].reduction > 1:
                reduction = self.config[self.task].reduction
                seg_len, n_classes = seg_label.shape
                seg_label = np.stack(
                    arrays=[
                        np.mean(
                            seg_data[reduction * idx : reduction * (idx + 1)],
                            axis=0,
                            keepdims=True,
                        ).astype(int)
                        for idx in range(seg_len // reduction)
                    ],
                    axis=0,
                ).squeeze(axis=1)
            seg_data, _ = self.seg_ppm(seg_data, self.config.fs)
            if self.task == [
                "rhythm_segmentation",
                "af_event",  # segmentation of AF events
            ]:
                weight_mask = generate_weight_mask(
                    target_mask=seg_label.squeeze(-1),
                    fg_weight=2,
                    fs=self.config.fs,
                    reduction=self.config[self.task].reduction,
                    radius=0.8,
                    boundary_weight=5,
                )[..., np.newaxis]
                return seg_data, seg_label, weight_mask
            return seg_data, seg_label, None
        elif self.task in [
            "rr_lstm",
        ]:
            seq_name = self.files[index]
            rec = seq_name.split("_")[1]
            rr_seq_path = self.file_dirs[rec] / f"{seq_name}.{self.file_ext}"
            rr_seq = loadmat(str(rr_seq_path))
            rr_seq["rr"] = rr_seq["rr"].reshape((self.seglen, 1))
            rr_seq["label"] = rr_seq["label"].reshape((self.seglen, self.n_classes))
            # map values of `rr_seq["label"]` to 0, 1 according to `self.reader.rhythm_types_map`
            rr_seq["label"][
                np.where(rr_seq["label"] == self.rhythm_types_map["AFIB"])
            ] = 1
            rr_seq["label"][
                np.where(rr_seq["label"] != self.rhythm_types_map["AFIB"])
            ] = 0
            weight_mask = generate_weight_mask(
                target_mask=rr_seq["label"].squeeze(-1),
                fg_weight=2,
                fs=1 / 0.8,
                reduction=1,
                radius=2,
                boundary_weight=5,
            )[..., np.newaxis]
            return rr_seq["rr"], rr_seq["label"], weight_mask
        else:
            raise NotImplementedError(
                f"data generator for task \042{self.task}\042 not implemented"
            )

    def __len__(self) -> int:
        return len(self.files)

    def extra_repr_keys(self) -> List[str]:
        return [
            "task",
            "reader",
            "ppm",
        ]



"""
"""

from numbers import Real
from typing import Any, Iterable, List, Optional, Sequence, Tuple

import numpy as np
from torch import Tensor

from ..cfg import DEFAULTS
from ..utils.utils_signal_t import normalize as normalize_t
from .base import Augmenter

__all__ = [
    "RandomRenormalize",
]


class RandomRenormalize(Augmenter):
    """Randomly re-normalize the ECG tensor,
    using the Z-score normalization method.

    Parameters
    ----------
    mean : array_like, default ``[-0.05, 0.1]``
        Range of mean value of the re-normalized signal, of shape ``(2,)``;
        or range of mean values for each lead of the re-normalized signal,
        of shape ``(lead, 2)``.
    std : array_like, default ``[0.08, 0.32]``
        Range of standard deviation of the re-normalized signal, of shape ``(2,)``;
        or range of standard deviations for each lead of the re-normalized signal,
        of shape ``(lead, 2)``.
    per_channel : bool, default False
        If True, re-normalization will be done per channel.
    prob : float, default 0.5
        Probability of applying the random re-normalization augmenter.
    inplace : bool, default True
        Whether to apply the random re-normalization augmenter in-place.
    kwargs : dict, optional
        Additional keyword arguments.

    Examples
    --------
    .. code-block:: python

        rrn = RandomRenormalize()
        sig = torch.randn(32, 12, 5000)
        sig, _ = rrn(sig, None)

    """

    __name__ = "RandomRenormalize"

    def __init__(
        self,
        mean: Iterable[Real] = [-0.05, 0.1],
        std: Iterable[Real] = [0.08, 0.32],
        per_channel: bool = False,
        prob: float = 0.5,
        inplace: bool = True,
        **kwargs: Any
    ) -> None:
        super().__init__()
        self.mean = np.array(mean)
        self.mean_mean = self.mean.mean(axis=-1, keepdims=True)
        self.mean_scale = (self.mean[..., -1] - self.mean_mean) * 0.3
        self.std = np.array(std)
        self.std_mean = self.std.mean(axis=-1, keepdims=True)
        self.std_scale = (self.std[..., -1] - self.std_mean) * 0.3
        self.per_channel = per_channel
        if not self.per_channel:
            assert self.mean.ndim == 1 and self.std.ndim == 1
        self.prob = prob
        self.inplace = inplace

    def forward(
        self,
        sig: Tensor,
        label: Optional[Tensor],
        *extra_tensors: Sequence[Tensor],
        **kwargs: Any
    ) -> Tuple[Tensor, ...]:
        """Forward function of the RandomRenormalize augmenter.

        Parameters
        ----------
        sig : torch.Tensor
            The input ECG tensor, of shape ``(batch, lead, siglen)``.
        label : torch.Tensor, optional
            The input ECG label tensor.
            Not used, but kept for compatibility with other augmenters.
        extra_tensors : Sequence[torch.Tensor], optional,
            Not used, but kept for consistency with other augmenters.
        kwargs : dict, optional
            Not used, but kept for consistency with other augmenters.

        Returns
        -------
        sig : torch.Tensor
            The randomly re-normalized ECG tensor.
        label : torch.Tensor
            The label tensor of the augmented ECGs, unchanged.
        extra_tensors: Sequence[torch.Tensor], optional,
            Unchanged extra tensors.

        """
        batch, lead, siglen = sig.shape
        if self.mean.ndim == 2:
            assert self.mean.shape[0] == lead
        if self.std.ndim == 2:
            assert self.std.shape[0] == lead
        if not self.inplace:
            sig = sig.clone()
        if self.prob == 0:
            return (sig, label, *extra_tensors)
        indices = self.get_indices(self.prob, pop_size=batch)
        if self.per_channel:
            mean = DEFAULTS.RNG.normal(
                self.mean_mean, self.mean_scale, size=(len(indices), lead, 1)
            )
            std = DEFAULTS.RNG.normal(
                self.std_mean, self.std_scale, size=(len(indices), lead, 1)
            )
        else:
            mean = DEFAULTS.RNG.normal(
                self.mean_mean, self.mean_scale, size=(len(indices), 1, 1)
            )
            std = DEFAULTS.RNG.normal(
                self.std_mean, self.std_scale, size=(len(indices), 1, 1)
            )
        sig[indices, ...] = normalize_t(
            sig[indices, ...],
            method="z-score",
            mean=mean,
            std=std,
            per_channel=self.per_channel,
            inplace=True,
        )
        return (sig, label, *extra_tensors)

    def extra_repr_keys(self) -> List[str]:
        return [
            "mean",
            "std",
            "per_channel",
            "prob",
            "inplace",
        ] + super().extra_repr_keys()



"""
MobileNets, from V1 to V3

References
----------
1. Howard, A. G., Zhu, M., Chen, B., Kalenichenko, D., Wang, W., Weyand, T., ... & Adam, H. (2017). Mobilenets: Efficient convolutional neural networks for mobile vision applications. arXiv preprint arXiv:1704.04861.
2. Sandler, M., Howard, A., Zhu, M., Zhmoginov, A., & Chen, L. C. (2018). Mobilenetv2: Inverted residuals and linear bottlenecks. In Proceedings of the IEEE conference on computer vision and pattern recognition (pp. 4510-4520).
3. Howard, A., Sandler, M., Chu, G., Chen, L. C., Chen, B., Tan, M., ... & Adam, H. (2019). Searching for mobilenetv3. In Proceedings of the IEEE International Conference on Computer Vision (pp. 1314-1324).

"""

import textwrap
from copy import deepcopy
from itertools import repeat
from numbers import Real
from typing import Any, Optional, Sequence, Union, List

from torch import Tensor, nn
from deprecate_kwargs import deprecate_kwargs

from ...cfg import CFG
from ...models._nets import (
    Conv_Bn_Activation,
    Initializers,
    MultiConv,
    make_attention_layer,
)
from ...utils.misc import add_docstring, CitationMixin
from ...utils.utils_nn import (
    SizeMixin,
    compute_sequential_output_shape,
    compute_sequential_output_shape_docstring,
)


__all__ = [
    "MobileNetV1",
    "MobileNetV2",
    "MobileNetV3",
]


_DEFAULT_CONV_CONFIGS_MobileNetV1 = CFG(
    ordering="cba",
    conv_type="separable",
    batch_norm=True,
    activation="relu6",
    kw_activation={"inplace": True},
    # kernel_initializer="he_normal",
    # kw_initializer={},
)


class MobileNetSeparableConv(nn.Sequential, SizeMixin):
    """Separable convolution.

    Similar to :class:`~torch_ecg._nets.SeparableConv`,
    the difference is that there are normalization and activation
    between depthwise conv and pointwise conv.

    Parameters
    ----------
    in_channels : int
        Number of channels in the input tensor.
    out_channels : int
        Number of channels produced by the convolution.
    kernel_size : int
        Size (length) of the convolution kernel.
    stride : int
        Stride (subsample length) of the convolution.
    padding : int, optional
        Zero-padding added to both sides of the input.
    dilation: int, default 1
        Spacing between the kernel points.
    groups : int, default 1
        Connection pattern (of channels) of the inputs and outputs.
    norm : bool or str or torch.nn.Module, default True
        (batch) normalization, or other normalizations,
        e.g. group normalization.
        (the name of) the Module itself or
        (if is bool) whether or not to use :class:`torch.nn.BatchNorm1d`.
    activation : str or torch.nn.Module, default "relu6"
        Name or Module of the activation.
        If is str, can be one of
        "mish", "swish", "relu", "leaky", "leaky_relu",
        "linear", "hardswish", "relu6".
        "linear" is equivalent to ``activation=None``.
    kernel_initializer : str or callable, optional
        A function to initialize kernel weights of the convolution,
        or name or the initialzer,
        can be one of the keys of ``Initializers``.
    bias : bool, default True
        If True, adds a learnable bias to the output.
    depth_multiplier : int, default 1
        Multiplier of the number of output channels
        of the depthwise convolution.
    width_multiplier : float, default 1.0
        Multiplier of the number of output channels
        of the pointwise convolution.
    kwargs : dict, optional,
        Extra parameters, including :attr:`ordering`, etc.

    """

    __name__ = "MobileNetSeparableConv"

    @deprecate_kwargs([["norm", "batch_norm"]])
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int,
        stride: int,
        padding: Optional[int] = None,
        dilation: int = 1,
        groups: int = 1,
        batch_norm: Union[bool, str, nn.Module] = True,
        activation: Optional[Union[str, nn.Module]] = "relu6",
        kernel_initializer: Optional[Union[str, callable]] = None,
        bias: bool = True,
        depth_multiplier: int = 1,
        width_multiplier: float = 1.0,
        **kwargs: Any,
    ) -> None:
        super().__init__()
        self.__in_channels = in_channels
        self.__out_channels = out_channels
        self.__kernel_size = kernel_size
        self.__stride = stride
        self.__dilation = dilation
        if padding is None:
            # "same" padding
            self.__padding = (self.__dilation * (self.__kernel_size - 1)) // 2
        elif isinstance(padding, int):
            self.__padding = padding
        self.__groups = groups
        self.__bias = bias
        ordering = kwargs.get("ordering", _DEFAULT_CONV_CONFIGS_MobileNetV1["ordering"])
        kw_initializer = kwargs.get("kw_initializer", {})
        self.__depth_multiplier = depth_multiplier
        dc_out_channels = int(self.__in_channels * self.__depth_multiplier)
        assert (
            dc_out_channels % self.__in_channels == 0
        ), f"depth_multiplier (input is {self.__depth_multiplier}) should be positive integers"
        self.__width_multiplier = width_multiplier
        self.__out_channels = int(self.__width_multiplier * self.__out_channels)
        assert (
            self.__out_channels % self.__groups == 0
        ), f"width_multiplier (input is {self.__width_multiplier}) makes `out_channels` not divisible by `groups` (= {self.__groups})"

        self.add_module(
            "depthwise_conv",
            Conv_Bn_Activation(
                in_channels=self.__in_channels,
                out_channels=dc_out_channels,
                kernel_size=self.__kernel_size,
                stride=self.__stride,
                padding=self.__padding,
                dilation=self.__dilation,
                groups=self.__in_channels,
                bias=self.__bias,
                norm=batch_norm,
                activation=activation,
                ordering=ordering,
            ),
        )
        self.add_module(
            "pointwise_conv",
            Conv_Bn_Activation(
                in_channels=dc_out_channels,
                out_channels=self.__out_channels,
                groups=self.__groups,
                bias=self.__bias,
                kernel_size=1,
                stride=1,
                padding=0,
                dilation=1,
                norm=batch_norm,
                activation=activation,
                ordering=ordering,
            ),
        )

        if kernel_initializer:
            if callable(kernel_initializer):
                for module in self:
                    kernel_initializer(module.weight)
            elif (
                isinstance(kernel_initializer, str)
                and kernel_initializer.lower() in Initializers.keys()
            ):
                for module in self:
                    Initializers[kernel_initializer.lower()](
                        module.weight, **kw_initializer
                    )
            else:  # TODO: add more initializers
                raise ValueError(f"initializer `{kernel_initializer}` not supported")

    @add_docstring(
        textwrap.indent(compute_sequential_output_shape_docstring, " " * 4),
        mode="append",
    )
    def compute_output_shape(
        self, seq_len: Optional[int] = None, batch_size: Optional[int] = None
    ) -> Sequence[Union[int, None]]:
        """Compute the output shape of the module."""
        return compute_sequential_output_shape(self, seq_len, batch_size)


class MobileNetV1(nn.Sequential, SizeMixin, CitationMixin):
    """MobileNet V1.

    Similar to :class:`~torch_ecg.models.Xception`, but without skip connections.
    Separable convolutions are slightly different too.

    normal conv
    --> entry flow (separable convs, down sample and double channels every other conv)
    --> middle flow (separable convs, no down sampling, stationary number of channels)
    --> exit flow (separable convs, down sample and double channels at each conv).

    MobileNet V1 is a light-weight model intended for mobile devices,
    originally proposed in [#v1]_.

    Parameters
    ----------
    in_channels : int
        Number of channels in the input.
    config : dict
        Other hyper-parameters of the Module, ref. corresponding config file.
        key word arguments that have to be set in 3 sub-dict,
        namely in "entry_flow", "middle_flow", and "exit_flow", including

            - out_channels: int,
              number of channels of the output.
            - kernel_size: int,
              kernel size of down sampling.
              If not specified, defaults to `down_scale`.
            - groups: int,
              connection pattern (of channels) of the inputs and outputs.
            - padding: int,
              zero-padding added to both sides of the input.
            - batch_norm: bool or Module,
              batch normalization, the Module itself
              or (if is bool) whether or not to use :class:`torch.nn.BatchNorm1d`.

    References
    ----------
    .. [#v1] Howard, A. G., Zhu, M., Chen, B., Kalenichenko, D., Wang, W., Weyand, T., ... & Adam, H. (2017).
           Mobilenets: Efficient convolutional neural networks for mobile vision applications.
           arXiv preprint arXiv:1704.04861.

    """

    __name__ = "MobileNetV1"

    def __init__(self, in_channels: int, **config) -> None:
        super().__init__()
        self.__in_channels = in_channels
        self.config = CFG(deepcopy(config))

        if isinstance(self.config.init_num_filters, int):
            init_convs = Conv_Bn_Activation(
                in_channels=self.__in_channels,
                out_channels=self.config.init_num_filters,
                kernel_size=self.config.init_filter_lengths,
                stride=self.config.init_subsample_lengths,
                groups=self.config.groups,
                norm=self.config.get("norm", self.config.get("batch_norm")),
                activation=self.config.activation,
                bias=self.config.bias,
                width_multiplier=self.config.width_multiplier,
            )
        else:
            init_convs = MultiConv(
                in_channels=self.__in_channels,
                out_channels=self.config.init_num_filters,
                filter_lengths=self.config.init_filter_lengths,
                subsample_lengths=self.config.init_subsample_lengths,
                groups=self.config.groups,
                norm=self.config.batch_norm,
                activation=self.config.activation,
                bias=self.config.bias,
                width_multiplier=self.config.width_multiplier,
            )
        self.add_module(
            "init_convs",
            init_convs,
        )

        _, entry_flow_in_channels, _ = self.init_convs.compute_output_shape()
        entry_flow = self._generate_flow(
            in_channels=entry_flow_in_channels,
            depth_multiplier=self.config.depth_multiplier,
            width_multiplier=self.config.width_multiplier,
            **self.config.entry_flow,
        )
        self.add_module("entry_flow", entry_flow)

        _, middle_flow_in_channels, _ = entry_flow[-1].compute_output_shape()
        middle_flow = self._generate_flow(
            in_channels=middle_flow_in_channels,
            depth_multiplier=self.config.depth_multiplier,
            width_multiplier=self.config.width_multiplier,
            **self.config.middle_flow,
        )
        self.add_module("middle_flow", middle_flow)

        _, exit_flow_in_channels, _ = middle_flow[-1].compute_output_shape()
        exit_flow = self._generate_flow(
            in_channels=exit_flow_in_channels,
            depth_multiplier=self.config.depth_multiplier,
            width_multiplier=self.config.width_multiplier,
            **self.config.exit_flow,
        )
        self.add_module(
            "exit_flow",
            exit_flow,
        )

    def _generate_flow(
        self,
        in_channels: int,
        out_channels: Sequence[int],
        filter_lengths: Union[Sequence[int], int],
        subsample_lengths: Union[Sequence[int], int] = 1,
        dilations: Union[Sequence[int], int] = 1,
        groups: int = 1,
        batch_norm: Union[bool, str, nn.Module] = True,
        activation: Optional[Union[str, nn.Module]] = "relu6",
        depth_multiplier: int = 1,
        width_multiplier: float = 1.0,
        ordering: str = "cba",
    ) -> nn.Sequential:
        """Generate a sequential flow of modules.

        Parameters
        ----------
        in_channels : int
            Number of channels in the input.
        out_channels : Sequence[int]
            Number of channels produced by the convolutional layers.
        filter_lengths : int or Sequence[int]
            Length(s) of the filters (kernel size).
        subsample_lengths  int or Sequence[int]
            Subsample length(s) (stride(s)) of the convolutions.
        dilations : int or Sequence[int], default 1,
            Spacing between the kernel points of (each) convolutional layer.
        groups : int, default 1
            Connection pattern (of channels) of the inputs and outputs.
        batch_norm : bool or str or torch.nn.Module, default True
            (batch) normalization, or other normalizations, e.g. group normalization.
            (the name of) the Module itself,
            or (if is bool) whether or not to use :class:`torch.nn.BatchNorm1d`.
        activation : str or torch.nn.Module, default "relu6"
            Name or Module of the activation.
        depth_multiplier : int, default 1
            Multiplier of the number of output channels of the depthwise convolution.
        width_multiplier : float, default 1.0
            Multiplier of the number of output channels of the pointwise convolution

        Returns
        -------
        flow : torch.nn.Sequential
            The sequential flow of consecutive separable convolutions,
            each followed by batch normalization and activation.

        """
        n_convs = len(out_channels)
        _filter_lengths = (
            list(repeat(filter_lengths, n_convs))
            if isinstance(filter_lengths, int)
            else filter_lengths
        )
        _subsample_lengths = (
            list(repeat(subsample_lengths, n_convs))
            if isinstance(subsample_lengths, int)
            else subsample_lengths
        )
        _dilations = (
            list(repeat(dilations, n_convs))
            if isinstance(dilations, int)
            else dilations
        )
        assert (
            n_convs
            == len(_filter_lengths)
            == len(_subsample_lengths)
            == len(_dilations)
        )
        ic = in_channels
        flow = nn.Sequential()
        for idx, (oc, fl, sl, dl) in enumerate(
            zip(out_channels, _filter_lengths, _subsample_lengths, _dilations)
        ):
            sc_layer = MobileNetSeparableConv(
                in_channels=ic,
                out_channels=oc,
                kernel_size=fl,
                stride=sl,
                dilation=dl,
                groups=groups,
                batch_norm=batch_norm,
                activation=activation,
                depth_multiplier=depth_multiplier,
                width_multiplier=width_multiplier,
                ordering=ordering,
            )
            flow.add_module(
                f"separable_conv_{idx}",
                sc_layer,
            )
            _, ic, _ = sc_layer.compute_output_shape()
        return flow

    def compute_output_shape(
        self, seq_len: Optional[int] = None, batch_size: Optional[int] = None
    ) -> Sequence[Union[int, None]]:
        """Compute the output shape of the model.

        Parameters
        ----------
        seq_len : int, optional
            Length of the input tensor.
        batch_size : int, optional,
            The batch size of the input tensor.

        Returns
        -------
        output_shape : sequence
            Output shape of the model.

        """
        _seq_len = seq_len
        _, _, _seq_len = self.init_convs.compute_output_shape(_seq_len, batch_size)
        for module in self.entry_flow:
            output_shape = module.compute_output_shape(_seq_len, batch_size)
            _, _, _seq_len = output_shape
        for module in self.middle_flow:
            output_shape = module.compute_output_shape(_seq_len, batch_size)
            _, _, _seq_len = output_shape
        for module in self.exit_flow:
            output_shape = module.compute_output_shape(_seq_len, batch_size)
            _, _, _seq_len = output_shape
        return output_shape

    @property
    def in_channels(self) -> int:
        return self.__in_channels

    @property
    def doi(self) -> List[str]:
        return list(set(self.config.get("doi", []) + ["10.48550/ARXIV.1704.04861"]))


class InvertedResidual(nn.Module, SizeMixin):
    """Inverted residual block.

    expansion (via pointwise conv) --> depthwise conv --> pointwise conv (without activation) ---> output
        |                                                                                      |
        |----------------------- shortcut (only under certain condition) ----------------------|

    Parameters
    ----------
    in_channels : int
        Number of channels in the input tensor.
    out_channels : int
        Number of channels produced by the convolution.
    expansion : float
        Expansion of the first pointwise convolution.
    filter_length : int
        Size (length) of the middle depthwise convolution kernel.
    stride: int
        Stride (subsample length) of the middle depthwise convolution.
    groups : int, default 1
        Connection pattern (of channels) of the inputs and outputs.
    dilation : int, default 1
        Spacing between the kernel points of (each) convolutional layer.
    norm : bool or str or torch.nn.Module, default True
        (batch) normalization, or other normalizations, e.g. group normalization.
        (the name of) the Module itself,
        or (if is bool) whether or not to use :class:`torch.nn.BatchNorm1d`.
    activation : str or torch.nn.Module, default "relu6"
        Name or Module of the activation, which is applied after convolutions,
        except for the last pointwise convolution.
    width_multiplier : float, default 1.0
        Multiplier of the number of output channels of the pointwise convolution.
    attn : dict, optional
        Attention mechanism for the neck conv layers.
        If is None, no attention mechanism is used.
        Keys:

            - "name": str, can be "se", "gc", "nl" (alias "nonlocal", "non-local"), etc.
            - "pos": int, position of the attention mechanism,
              other keys are specific to the attention mechanism.

    """

    __name__ = "InvertedResidual"

    @deprecate_kwargs([["norm", "batch_norm"]])
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        expansion: float,
        filter_length: int,
        stride: int,
        groups: int = 1,
        dilation: int = 1,
        batch_norm: Union[bool, str, nn.Module] = True,
        activation: Optional[Union[str, nn.Module]] = "relu6",
        width_multiplier: float = 1.0,
        attn: Optional[CFG] = None,
    ) -> None:
        super().__init__()
        self.__in_channels = in_channels
        self.__out_channels = out_channels
        self.__expansion = expansion
        self.__mid_channels = int(round(self.__expansion * self.__in_channels))
        self.__filter_length = filter_length
        self.__stride = stride
        self.__groups = groups
        assert self.__mid_channels % self.__groups == 0, (
            f"expansion ratio (input is {self.__expansion}) makes mid-channels "
            f"(= {self.__mid_channels}) not divisible by `groups` (={self.__groups})"
        )
        self.__dilation = dilation
        if self.__dilation > 1:
            self.__stride = 1
        self.__width_multiplier = width_multiplier
        self.__attn = attn
        if self.__attn:
            self.__attn = CFG(self.__attn)

        self.main_stream = nn.Sequential()
        conv_in_channels = self.__in_channels
        current_pos = 0
        current_pos = self._add_attn_layer_if_needed(conv_in_channels, current_pos)
        if self.__expansion != 1:
            # expand, pointwise
            expansion = Conv_Bn_Activation(
                conv_in_channels,
                self.__mid_channels,
                kernel_size=1,
                stride=1,
                groups=self.__groups,
                norm=batch_norm,
                activation=activation,
                # width_multiplier=width_multiplier,
            )
            self.main_stream.add_module(
                "expansion",
                expansion,
            )
            current_pos += 1
            current_pos = self._add_attn_layer_if_needed(conv_in_channels, current_pos)
            _, conv_in_channels, _ = expansion.compute_output_shape()
        # depthwise convolution
        dw = Conv_Bn_Activation(
            conv_in_channels,
            conv_in_channels,
            kernel_size=self.__filter_length,
            stride=self.__stride,
            groups=conv_in_channels,
            dilation=self.__dilation,
            norm=batch_norm,
            activation=activation,
            # width_multiplier=width_multiplier,
        )
        self.main_stream.add_module(
            "depthwise_conv",
            dw,
        )
        current_pos += 1
        current_pos = self._add_attn_layer_if_needed(conv_in_channels, current_pos)
        _, conv_in_channels, _ = dw.compute_output_shape()
        # pointwise conv without non-linearity (activation)
        pw_linear = Conv_Bn_Activation(
            conv_in_channels,
            self.__out_channels,
            kernel_size=1,
            stride=1,
            groups=self.__groups,
            bias=False,
            norm=batch_norm,
            activation=None,
            width_multiplier=self.__width_multiplier,
        )
        self.main_stream.add_module(
            "pointwise_conv",
            pw_linear,
        )
        current_pos += 1
        current_pos = self._add_attn_layer_if_needed(conv_in_channels, current_pos)
        _, self.__out_channels, _ = pw_linear.compute_output_shape()

    def _add_attn_layer_if_needed(self, in_channels: int, current_pos: int) -> int:
        """Add attention layer.

        Add attention layer at the position specified by `self.__attn.pos`.

        Parameters
        ----------
        in_channels : int
            Number of channels in the input tensor.
        current_pos : int
            Position of the current layer.

        Returns
        -------
        int
            Position of the next layer.

        """
        if self.__attn and self.__attn["pos"] == current_pos:
            self.main_stream.add_module(
                self.__attn["name"],
                make_attention_layer(in_channels, **self.__attn),
            )
            return current_pos + 1
        return current_pos

    def forward(self, input: Tensor) -> Tensor:
        """Forward pass.

        Parameters
        ----------
        input : torch.Tensor
            Input tensor,
            of shape ``(batch_size, n_channels, seq_len)``.

        Returns
        -------
        out : torch.Tensor
            Output tensor,
            of shape ``(batch_size, n_channels, seq_len)``.

        """
        out = self.main_stream(input)

        if self.__stride == 1 and self.__in_channels == self.__out_channels:
            # NOTE the condition that skip connections are done
            # which is different from ResNet
            out = out + input

        return out

    def compute_output_shape(
        self, seq_len: Optional[int] = None, batch_size: Optional[int] = None
    ) -> Sequence[Union[int, None]]:
        """Compute the output shape of the Module.

        Parameters
        ----------
        seq_len : int, optional
            Length of the input tensor.
        batch_size : int, optional
            The batch size of the input tensor.

        Returns
        -------
        output_shape : sequence
            Output shape of the Module.

        """
        _seq_len = seq_len
        for module in self.main_stream:
            if isinstance(module, nn.Dropout):
                continue
            output_shape = module.compute_output_shape(_seq_len, batch_size)
            _, _, _seq_len = output_shape
        return output_shape


class MobileNetV2(nn.Sequential, SizeMixin, CitationMixin):
    """MobileNet V2.

    MobileNet V2 is an upgraded version of MobileNet V1, originally proposed in
    [#v2]_. It uses inverted residual blocks instead of the original residual
    blocks. Torchvision's implementation [#v2_pt] and Keras' implementation [#v2_keras]_
    are used as references.

    Parameters
    ----------
    in_channels : int
        Number of channels in the input signal tensor.
    config : dict
        Other hyper-parameters of the Module, ref. corr. config file
        keyword arguments that have to be set are as follows:

        - groups: int,
          number of groups in the pointwise convolutional layer(s).
        - norm: bool or str or Module,
          normalization layer.
        - activation: str or Module,
          activation layer.
        - bias: bool,
          whether to use bias in the convolutional layer(s).
        - width_multiplier: float,
          multiplier of the number of output channels of the pointwise convolution.
        - stem: CFG,
          config of the stem block, with the following keys:

            - num_filters: int or Sequence[int],
              number of filters in the first convolutional layer(s).
            - filter_lengths: int or Sequence[int],
              filter lengths (kernel sizes) in the first convolutional layer(s).
            - subsample_lengths: int or Sequence[int],
              subsample lengths (strides) in the first convolutional layer(s).

        - inv_res: CFG,
          Config of the inverted residual blocks, with the following keys:

            - expansions: Sequence[int],
              expansion ratios of the inverted residual blocks.
            - out_channels: Sequence[int],
              number of output channels in each block.
            - n_blocks: Sequence[int],
              number of inverted residual blocks.
            - strides: Sequence[int],
              strides of the inverted residual blocks.
            - filter_lengths: Sequence[int],
              filter lengths (kernel sizes) in each block.

        - exit_flow: CFG,
          Config of the exit flow blocks, with the following keys:

            - num_filters: int or Sequence[int],
              number of filters in the final convolutional layer(s).
            - filter_lengths: int or Sequence[int],
              filter lengths (kernel sizes) in the final convolutional layer(s).
            - subsample_lengths: int or Sequence[int],
              subsample lengths (strides) in the final convolutional layer(s).

    References
    ----------
    .. [#v2] Sandler, M., Howard, A., Zhu, M., Zhmoginov, A., & Chen, L. C. (2018).
             Mobilenetv2: Inverted residuals and linear bottlenecks. In Proceedings of the
             IEEE conference on computer vision and pattern recognition (pp. 4510-4520).
    .. [#v2_pt] https://github.com/pytorch/vision/blob/master/torchvision/models/mobilenetv2.py
    .. [#v2_keras] https://github.com/keras-team/keras-applications/blob/master/keras_applications/mobilenet_v2.py

    """

    __name__ = "MobileNetV2"

    def __init__(self, in_channels: int, **config: CFG) -> None:
        super().__init__()
        self.__in_channels = in_channels
        self.config = CFG(deepcopy(config))

        # stem
        if isinstance(self.config.stem.num_filters, int):
            stem = Conv_Bn_Activation(
                in_channels=self.__in_channels,
                out_channels=self.config.stem.num_filters,
                kernel_size=self.config.stem.filter_lengths,
                stride=self.config.stem.subsample_lengths,
                groups=self.config.groups,
                norm=self.config.get("norm", self.config.get("batch_norm")),
                activation=self.config.activation,
                bias=self.config.bias,
                width_multiplier=self.config.width_multiplier,
            )
        else:
            stem = MultiConv(
                in_channels=self.__in_channels,
                out_channels=self.config.stem.num_filters,
                filter_lengths=self.config.stem.filter_lengths,
                subsample_lengths=self.config.stem.subsample_lengths,
                groups=self.config.groups,
                norm=self.config.get("norm", self.config.get("batch_norm")),
                activation=self.config.activation,
                bias=self.config.bias,
                width_multiplier=self.config.width_multiplier,
            )
        self.add_module(
            "stem",
            stem,
        )

        # inverted residual blocks
        inv_res_cfg = zip(
            self.config.inv_res.expansions,
            self.config.inv_res.out_channels,
            self.config.inv_res.n_blocks,
            self.config.inv_res.strides,
            self.config.inv_res.filter_lengths,
        )
        _, inv_res_in_channels, _ = stem.compute_output_shape()
        idx = 0
        for t, c, n, s, k in inv_res_cfg:
            # t: expansion
            # c: output channels
            # n: number of blocks
            # s: stride
            # k: kernel size
            for i in range(n):
                inv_res_blk = InvertedResidual(
                    inv_res_in_channels,
                    out_channels=c,
                    expansion=t,
                    filter_length=k,
                    stride=s if i == 0 else 1,
                    groups=self.config.groups,
                    norm=self.config.get("norm", self.config.get("batch_norm")),
                    activation=self.config.activation,
                    width_multiplier=self.config.width_multiplier,
                )
                self.add_module(
                    f"inv_res_{idx}",
                    inv_res_blk,
                )
                _, inv_res_in_channels, _ = inv_res_blk.compute_output_shape()
                idx += 1

        # exit_flow
        # no alpha applied to last conv as stated in the paper
        if isinstance(self.config.exit_flow.num_filters, int):
            exit_flow = Conv_Bn_Activation(
                in_channels=inv_res_in_channels,
                out_channels=self.config.exit_flow.num_filters,
                kernel_size=self.config.exit_flow.filter_lengths,
                stride=self.config.exit_flow.subsample_lengths,
                groups=self.config.groups,
                norm=self.config.get("norm", self.config.get("batch_norm")),
                activation=self.config.activation,
                bias=self.config.bias,
                width_multiplier=max(1.0, self.config.width_multiplier),
            )
        else:
            exit_flow = MultiConv(
                in_channels=self.__in_channels,
                out_channels=self.config.exit_flow.num_filters,
                filter_lengths=self.config.exit_flow.filter_lengths,
                subsample_lengths=self.config.exit_flow.subsample_length,
                groups=self.config.groups,
                norm=self.config.get("norm", self.config.get("batch_norm")),
                activation=self.config.activation,
                bias=self.config.bias,
                width_multiplier=max(1.0, self.config.width_multiplier),
            )
        self.add_module(
            "exit_flow",
            exit_flow,
        )

    @add_docstring(
        textwrap.indent(compute_sequential_output_shape_docstring, " " * 4),
        mode="append",
    )
    def compute_output_shape(
        self, seq_len: Optional[int] = None, batch_size: Optional[int] = None
    ) -> Sequence[Union[int, None]]:
        """Compute the output shape of the model."""
        return compute_sequential_output_shape(self, seq_len, batch_size)

    @property
    def in_channels(self) -> int:
        return self.__in_channels

    @property
    def doi(self) -> List[str]:
        return list(set(self.config.get("doi", []) + ["10.1109/cvpr.2018.00474"]))


class InvertedResidualBlock(nn.Sequential, SizeMixin):
    """Inverted residual block.

    Parameters
    ----------
    in_channels : int
        Number of input channels.
    n_blocks : int
        Number of inverted residual blocks.
    expansion : float or Sequence[float]
        Expansion ratios of the inverted residual blocks.
    filter_length : int or Sequence[int]
        Filter length of the depthwise convolution in the inverted residual blocks.
    stride : int or Sequence[int], optional
        Stride of the depthwise convolution in the inverted residual blocks,
        defaults to ``[2] + [1] * (n_blocks - 1)``.
    groups : int, default 1
        Number of groups in the expansion and pointwise convolution
        in the inverted residual blocks.
    dilation : int or Sequence[int], default 1
        Dilation of the depthwise convolution in the inverted residual blocks.
    norm : bool or str or torch.nn.Module, default True
        Normalization layer to use, defaults to batch normalization.
    activation : str or torch.nn.Module or Sequence[str] or Sequence[torch.nn.Module], default "relu"
        Activation function to use.
    width_multiplier : float or Sequence[float], default 1.0
        Width multiplier of the inverted residual blocks.
    out_channels : int or Sequence[int], optional
        Number of output channels of the inverted residual blocks,
        defaults to ``2 * in_channels``.
    attn : CFG or Sequence[CFG], optional
        Config of attention layer to use, defaults to None.

    """

    __name__ = "InvertedResidualBlock"

    @deprecate_kwargs([["norm", "batch_norm"]])
    def __init__(
        self,
        in_channels: int,
        n_blocks: int,
        expansion: Union[float, Sequence[float]],
        filter_length: Union[int, Sequence[int]],
        stride: Union[int, Sequence[int]] = None,
        groups: int = 1,
        dilation: Union[int, Sequence[int]] = 1,
        batch_norm: Union[bool, str, nn.Module] = True,
        activation: Optional[
            Union[str, nn.Module, Sequence[Union[str, nn.Module]]]
        ] = "relu",
        width_multiplier: Union[float, Sequence[float]] = 1.0,
        out_channels: Union[int, Sequence[int]] = None,
        attn: Optional[Union[CFG, Sequence[CFG]]] = None,
        **kwargs: Any,
    ) -> None:
        super().__init__()
        self.__in_channels = in_channels
        self.__n_blocks = n_blocks
        self.__expansion = expansion
        if isinstance(expansion, Real):
            self.__expansion = list(repeat(expansion, self.n_blocks))
        else:
            self.__expansion = expansion
        assert (
            len(self.__expansion) == self.n_blocks
        ), f"expansion must be an integer or a sequence of length {self.n_blocks}"
        if isinstance(filter_length, int):
            self.__filter_length = list(repeat(filter_length, self.n_blocks))
        else:
            self.__filter_length = filter_length
        assert (
            len(self.__filter_length) == self.n_blocks
        ), f"filter_length must be an integer or a sequence of length {self.n_blocks}"
        if stride is None:
            self.__stride = [2] + list(repeat(1, self.n_blocks - 1))
        elif isinstance(stride, int):
            self.__stride = list(repeat(stride, self.n_blocks))
        else:
            self.__stride = stride
        assert (
            len(self.__stride) == self.n_blocks
        ), f"stride must be an integer or a sequence of length {self.n_blocks}"
        self.__groups = groups
        if isinstance(dilation, int):
            self.__dilation = list(repeat(dilation, self.n_blocks))
        else:
            self.__dilation = dilation
        assert (
            len(self.__dilation) == self.n_blocks
        ), f"dilation must be an integer or a sequence of length {self.n_blocks}"
        self.__batch_norm = batch_norm
        if isinstance(activation, (str, nn.Module)):
            self.__activation = [deepcopy(activation) for _ in range(self.n_blocks)]
        else:
            self.__activation = activation
        assert (
            len(self.__activation) == self.n_blocks
        ), f"activation must be a string or Module or a sequence of length {self.n_blocks}"
        if isinstance(width_multiplier, float):
            self.__width_multiplier = list(repeat(width_multiplier, self.n_blocks))
        else:
            self.__width_multiplier = width_multiplier
        assert (
            len(self.__width_multiplier) == self.n_blocks
        ), f"width_multiplier must be a float or a sequence of length {self.n_blocks}"
        if out_channels is None:
            self.__out_channels = list(repeat(2 * in_channels, self.n_blocks))
        elif isinstance(out_channels, int):
            self.__out_channels = list(repeat(out_channels, self.n_blocks))
        else:
            self.__out_channels = out_channels
        assert (
            len(self.__out_channels) == self.n_blocks
        ), f"out_channels must be an integer or a sequence of length {self.n_blocks}"
        if attn is None or isinstance(attn, CFG):
            self.__attn = [deepcopy(attn) for _ in range(self.n_blocks)]
        else:
            self.__attn = attn
        assert (
            len(self.__attn) == self.n_blocks
        ), f"attn must be a CFG or a sequence of length {self.n_blocks}"

        ivt_res_in_channels = self.__in_channels
        for idx, exp in enumerate(self.__expansion):
            self.add_module(
                f"inv_res_{idx}",
                InvertedResidual(
                    in_channels=ivt_res_in_channels,
                    out_channels=self.__out_channels[idx],
                    expansion=exp,
                    filter_length=self.__filter_length[idx],
                    stride=self.__stride[idx] if idx == 0 else 1,
                    dilation=self.__dilation[idx],
                    groups=self.__groups,
                    norm=self.__batch_norm,
                    activation=self.__activation[idx],
                    width_multiplier=self.__width_multiplier[idx],
                    attn=self.__attn[idx],
                ),
            )
            ivt_res_in_channels = self.__out_channels[idx]

    @property
    def n_blocks(self) -> int:
        return self.__n_blocks

    @add_docstring(
        textwrap.indent(compute_sequential_output_shape_docstring, " " * 4),
        mode="append",
    )
    def compute_output_shape(
        self, seq_len: Optional[int] = None, batch_size: Optional[int] = None
    ) -> Sequence[Union[int, None]]:
        """Compute the output shape of the block."""
        return compute_sequential_output_shape(self, seq_len, batch_size)


class MobileNetV3_STEM(nn.Sequential, SizeMixin):
    """Stem block of MobileNet V3.

    Parameters
    ----------
    in_channels : int
        Number of channels in the input tensor.
    groups : int, default 1
        Number of groups in the expansion and pointwise convolution.
    bias : bool, default True
        Whether to use bias in the convolutional layer(s).
    norm : bool or str or torch.nn.Module, default True
        Normalization layer to use, defaults to batch normalization.
    activation : str or torch.nn.Module or Sequence[str] or Sequence[torch.nn.Module], default "relu"
        Activation function to use.
    width_multiplier : float or Sequence[float], default 1.0
        Width multiplier of the inverted residual blocks.
    config : CFG, optional
        Config of the stem block, with the following items:

            - num_filters: int or Sequence[int],
              number of filters in the first convolutional layer(s).
            - filter_lengths: int or Sequence[int],
              filter lengths (kernel sizes) in the first convolutional layer(s).
            - subsample_lengths: int or Sequence[int],
              subsample lengths (strides) in the first convolutional layer(s).

    """

    __name__ = "MobileNetV3_STEM"

    @deprecate_kwargs([["norm", "batch_norm"]])
    def __init__(
        self,
        in_channels: int,
        groups: int = 1,
        bias: bool = True,
        batch_norm: Union[bool, str, nn.Module] = True,
        activation: Optional[
            Union[str, nn.Module, Sequence[Union[str, nn.Module]]]
        ] = "relu",
        width_multiplier: Union[float, Sequence[float]] = 1.0,
        **config: CFG,
    ) -> None:
        super().__init__()
        self.__in_channels = in_channels
        self.config = CFG(deepcopy(config))
        if isinstance(self.config.num_filters, int):
            out_channels = self.config.num_filters
        else:
            out_channels = self.config.num_filters[0]
        self.add_module(
            "init_conv",
            Conv_Bn_Activation(
                in_channels=self.__in_channels,
                out_channels=out_channels,
                kernel_size=self.config.filter_lengths,
                stride=self.config.subsample_lengths,
                groups=groups,
                norm=batch_norm,
                activation=activation,
                bias=bias,
                width_multiplier=width_multiplier,
            ),
        )
        in_channels = out_channels
        if not isinstance(self.config.num_filters, int):
            for idx, out_channels in enumerate(self.config.num_filters[1:]):
                inv_res = InvertedResidual(
                    in_channels=in_channels,
                    out_channels=out_channels,
                    expansion=1.0,
                    filter_length=self.config.filter_lengths,
                    stride=1,
                    groups=groups,
                    dilation=1,
                    norm=batch_norm,
                    activation=activation,
                )
                self.add_module(f"inv_res_{idx}", inv_res)
                _, in_channels, _ = inv_res.compute_output_shape()

    @add_docstring(
        textwrap.indent(compute_sequential_output_shape_docstring, " " * 4),
        mode="append",
    )
    def compute_output_shape(
        self, seq_len: Optional[int] = None, batch_size: Optional[int] = None
    ) -> Sequence[Union[int, None]]:
        """Compute the output shape of the stem block."""
        return compute_sequential_output_shape(self, seq_len, batch_size)


class MobileNetV3(nn.Sequential, SizeMixin, CitationMixin):
    """MobileNet V3.

    MobileNet V3 [#v3]_ is an incremental improvement of MobileNet series.
    MobileNet V3 uses neural architecture search instead of hand-designed
    architectures to find the optimal network structure.
    MobileNet V3 has implementions in Torchvision [#v3_pt]_,
    which serves as a reference for this implementation for ECG tasks.

    Parameters
    ----------
    in_channels : int
        Number of channels in the input signal tensor.
    config : dict
        Other hyper-parameters of the Module, ref. corr. config file.
        Keyword arguments that must be set:

            - groups: int,
              number of groups in the convolutional layer(s) other than depthwise convolutions.
            - norm: bool or str or Module,
              normalization layer.
            - bias: bool,
              whether to use bias in the convolutional layer(s).
            - width_multiplier: float,
              multiplier of the number of output channels of the pointwise convolution.
            - stem: CFG,
              config of the stem block, with the following keys:

                - num_filters: int or Sequence[int],
                  number of filters in the first convolutional layer(s).
                - filter_lengths: int or Sequence[int],
                  filter lengths (kernel sizes) in the first convolutional layer(s).
                - subsample_lengths: int or Sequence[int],
                  subsample lengths (strides) in the first convolutional layer(s).

            - inv_res: CFG,
              config of the inverted residual blocks, with the following keys:

                - in_channels: Sequence[int],
                  number of input channels.
                - n_blocks: Sequence[int],
                  number of inverted residual blocks.
                - expansions: sequence of floats or sequence of sequence of floats,
                  expansion ratios of the inverted residual blocks.
                - filter_lengths: sequence of ints or sequence of sequence of ints,
                  filter length of the depthwise convolution in the inverted residual blocks.
                - stride: sequence of ints or sequence of sequence of ints, optional,
                  stride of the depthwise convolution in the inverted residual blocks,
                  defaults to ``[2] + [1] * (n_blocks - 1)``.
                - groups: int, default 1,
                  number of groups in the expansion and pointwise convolution
                  in the inverted residual blocks.
                - dilation: sequence of ints or sequence of sequence of ints, optional,
                  dilation of the depthwise convolution in the inverted residual blocks.
                - batch_norm: bool or str or nn.Module, default True,
                  normalization layer to use, defaults to batch normalization.
                - activation: str or nn.Module or sequence of str or torch.nn.Module,
                  activation function to use.
                - width_multiplier: float or sequence of floats, default 1.0,
                  width multiplier of the inverted residual blocks.
                - out_channels: sequence of ints or sequence of Sequence[int], optional,
                  number of output channels of the inverted residual blocks,
                  defaults to ``2 * in_channels``.
                - attn: sequence of CFG or sequence of sequence of CFG, optional,
                  config of attention layer to use, defaults to None.

            - exit_flow: CFG,
                config of the exit flow blocks, with the following keys:

                - num_filters: int or Sequence[int],
                  number of filters in the final convolutional layer(s).
                - filter_lengths: int or Sequence[int],
                  filter lengths (kernel sizes) in the final convolutional layer(s).
                - subsample_lengths: int or Sequence[int],
                  subsample lengths (strides) in the final convolutional layer(s).

    References
    ----------
    .. [#v3] Howard, A., Sandler, M., Chu, G., Chen, L. C., Chen, B., Tan, M., ... & Adam, H. (2019).
             Searching for mobilenetv3. In Proceedings of the IEEE International Conference
             on Computer Vision (pp. 1314-1324).
    .. [#v3_pt] https://github.com/pytorch/vision/blob/master/torchvision/models/mobilenetv3.py

    """

    __name__ = "MobileNetV3"

    def __init__(self, in_channels: int, **config: CFG) -> None:
        super().__init__()
        self.__in_channels = in_channels
        self.config = CFG(deepcopy(config))

        # stem
        self.add_module(
            "stem",
            MobileNetV3_STEM(
                in_channels=in_channels,
                groups=self.config.groups,
                bias=self.config.bias,
                norm=self.config.get("norm", self.config.get("batch_norm")),
                activation=self.config.activation,
                width_multiplier=self.config.width_multiplier,
                **self.config.stem,
            ),
        )

        # inverted residual blocks
        _, inv_res_in_channels, _ = self.stem.compute_output_shape()
        strides = self.config.inv_res.get(
            "strides", list(repeat(None, len(self.config.inv_res.n_blocks)))
        )
        out_channels = self.config.inv_res.get(
            "out_channels", list(repeat(None, len(self.config.inv_res.n_blocks)))
        )
        for idx, n_blocks in enumerate(self.config.inv_res.n_blocks):
            block = InvertedResidualBlock(
                in_channels=inv_res_in_channels,
                n_blocks=n_blocks,
                expansion=self.config.inv_res.expansions[idx],
                filter_length=self.config.inv_res.filter_lengths[idx],
                stride=strides[idx],
                groups=self.config.inv_res.groups,
                dilation=self.config.inv_res.dilations[idx],
                norm=self.config.get("norm", self.config.get("batch_norm")),
                activation=self.config.inv_res.activations[idx],
                bias=self.config.bias,
                width_multiplier=self.config.width_multiplier,
                out_channels=out_channels[idx],
                attn=self.config.inv_res.attns[idx],
            )
            self.add_module(
                f"block_{idx}",
                block,
            )
            _, inv_res_in_channels, _ = block.compute_output_shape()

        # exit_flow
        # no alpha applied to last conv as stated in the paper
        if isinstance(self.config.exit_flow.num_filters, int):
            exit_flow = Conv_Bn_Activation(
                in_channels=inv_res_in_channels,
                out_channels=self.config.exit_flow.num_filters,
                kernel_size=self.config.exit_flow.filter_lengths,
                stride=self.config.exit_flow.subsample_lengths,
                groups=self.config.groups,
                norm=self.config.get("norm", self.config.get("batch_norm")),
                activation=self.config.activation,
                bias=self.config.bias,
                width_multiplier=max(1.0, self.config.width_multiplier),
            )
        else:
            exit_flow = MultiConv(
                in_channels=self.__in_channels,
                out_channels=self.config.exit_flow.num_filters,
                filter_lengths=self.config.exit_flow.filter_lengths,
                subsample_lengths=self.config.exit_flow.subsample_length,
                groups=self.config.groups,
                norm=self.config.get("norm", self.config.get("batch_norm")),
                activation=self.config.activation,
                bias=self.config.bias,
                width_multiplier=max(1.0, self.config.width_multiplier),
            )
        self.add_module(
            "exit_flow",
            exit_flow,
        )

    @add_docstring(
        textwrap.indent(compute_sequential_output_shape_docstring, " " * 4),
        mode="append",
    )
    def compute_output_shape(
        self, seq_len: Optional[int] = None, batch_size: Optional[int] = None
    ) -> Sequence[Union[int, None]]:
        """Compute the output shape of the model."""
        return compute_sequential_output_shape(self, seq_len, batch_size)

    @property
    def in_channels(self) -> int:
        return self.__in_channels

    @property
    def doi(self) -> List[str]:
        return list(set(self.config.get("doi", []) + ["10.1109/iccv.2019.00140"]))



"""
configs of C(R)NN structure models, for ECG wave delineation
"""

from copy import deepcopy

from ..cfg import CFG
from ..utils.utils_nn import adjust_cnn_filter_lengths
from .attn import global_context, non_local
from .cnn import multi_scopic, multi_scopic_leadwise


__all__ = [
    "ECG_SEQ_LAB_NET_CONFIG",
]


# vanilla config, for delineation using single-lead ECG in corresponding papers
ECG_SEQ_LAB_NET_CONFIG = CFG()
ECG_SEQ_LAB_NET_CONFIG.fs = 500


ECG_SEQ_LAB_NET_CONFIG.cnn = CFG()
ECG_SEQ_LAB_NET_CONFIG.cnn.name = "multi_scopic"
ECG_SEQ_LAB_NET_CONFIG.cnn.multi_scopic = deepcopy(multi_scopic)
_base_num_filters = 4
ECG_SEQ_LAB_NET_CONFIG.cnn.multi_scopic.num_filters = [
    [
        _base_num_filters * 4,
        _base_num_filters * 8,
        _base_num_filters * 16,
    ],
    [
        _base_num_filters * 4,
        _base_num_filters * 8,
        _base_num_filters * 16,
    ],
    [
        _base_num_filters * 4,
        _base_num_filters * 8,
        _base_num_filters * 16,
    ],
]
ECG_SEQ_LAB_NET_CONFIG.cnn.multi_scopic = adjust_cnn_filter_lengths(
    ECG_SEQ_LAB_NET_CONFIG.cnn.multi_scopic, ECG_SEQ_LAB_NET_CONFIG.fs
)

ECG_SEQ_LAB_NET_CONFIG.cnn.multi_scopic_leadwise = deepcopy(multi_scopic_leadwise)
ECG_SEQ_LAB_NET_CONFIG.cnn.multi_scopic_leadwise = adjust_cnn_filter_lengths(
    ECG_SEQ_LAB_NET_CONFIG.cnn.multi_scopic_leadwise, ECG_SEQ_LAB_NET_CONFIG.fs
)


ECG_SEQ_LAB_NET_CONFIG.rnn = CFG()
ECG_SEQ_LAB_NET_CONFIG.rnn.name = "lstm"  # "none"
ECG_SEQ_LAB_NET_CONFIG.rnn.lstm = CFG()
ECG_SEQ_LAB_NET_CONFIG.rnn.lstm.hidden_sizes = [256, 256]
ECG_SEQ_LAB_NET_CONFIG.rnn.lstm.bias = True
ECG_SEQ_LAB_NET_CONFIG.rnn.lstm.retseq = True
ECG_SEQ_LAB_NET_CONFIG.rnn.lstm.dropouts = 0
ECG_SEQ_LAB_NET_CONFIG.rnn.lstm.bidirectional = True


ECG_SEQ_LAB_NET_CONFIG.attn = CFG()
ECG_SEQ_LAB_NET_CONFIG.attn.name = "se"  # "gc", "nl", "none"
ECG_SEQ_LAB_NET_CONFIG.attn.se = CFG()
ECG_SEQ_LAB_NET_CONFIG.attn.se.reduction = 8  # not including the last linear layer
ECG_SEQ_LAB_NET_CONFIG.attn.se.activation = "relu"
ECG_SEQ_LAB_NET_CONFIG.attn.se.kw_activation = CFG(inplace=True)
ECG_SEQ_LAB_NET_CONFIG.attn.se.bias = True
ECG_SEQ_LAB_NET_CONFIG.attn.se.kernel_initializer = "he_normal"
# ECG_SEQ_LAB_NET_CONFIG.attn.se.dropouts = [0.2, 0.0]
ECG_SEQ_LAB_NET_CONFIG.attn.gc = deepcopy(global_context)
ECG_SEQ_LAB_NET_CONFIG.attn.nl = deepcopy(non_local)


ECG_SEQ_LAB_NET_CONFIG.clf = CFG()
ECG_SEQ_LAB_NET_CONFIG.clf.out_channels = [
    256,
    64,
]  # not including the last linear layer
ECG_SEQ_LAB_NET_CONFIG.clf.activation = "mish"
ECG_SEQ_LAB_NET_CONFIG.clf.bias = True
ECG_SEQ_LAB_NET_CONFIG.clf.kernel_initializer = "he_normal"
ECG_SEQ_LAB_NET_CONFIG.clf.dropouts = [0.2, 0.2, 0.0]



"""BandPass filter preprocessor."""

from numbers import Real
from typing import Any, List, Optional, Tuple

import numpy as np

from .base import PreProcessor, preprocess_multi_lead_signal


__all__ = [
    "BandPass",
]


class BandPass(PreProcessor):
    """Bandpass filtering preprocessor.

    Parameters
    ----------
    lowcut : numbers.Real, optional
        Low cutoff frequency
    highcut : numbers.Real, optional
        High cutoff frequency.
    filter_type : {"butter", "fir"}, optional
        Type of the bandpass filter, default "butter".
    filter_order : int, optional
        Order of the bandpass filter.
    **kwargs : dict, optional
        Other arguments for :class:`PreProcessor`.

    Examples
    --------
    .. code-block:: python

        from torch_ecg.cfg import DEFAULTS
        sig = DEFAULTS.RNG.randn(1000)
        pp = BandPass(lowcut=0.5, highcut=45, filter_type="butter", filter_order=4)
        sig, _ = pp(sig, 500)

    """

    __name__ = "BandPass"

    def __init__(
        self,
        lowcut: Optional[Real] = 0.5,
        highcut: Optional[Real] = 45,
        filter_type: str = "butter",
        filter_order: Optional[int] = None,
        **kwargs: Any
    ) -> None:
        self.lowcut = lowcut
        self.highcut = highcut
        assert any(
            [self.lowcut is not None, self.highcut is not None]
        ), "At least one of lowcut and highcut should be set"
        if not self.lowcut:
            self.lowcut = 0
        if not self.highcut:
            self.highcut = float("inf")
        self.filter_type = filter_type
        self.filter_order = filter_order

    def apply(self, sig: np.ndarray, fs: int) -> Tuple[np.ndarray, int]:
        """Apply the preprocessor to `sig`.

        Parameters
        ----------
        sig : numpy.ndarray
            The ECG signal, can be
                - 1d array, which is a single-lead ECG;
                - 2d array, which is a multi-lead ECG of "lead_first" format;
                - 3d array, which is a tensor of several ECGs, of shape ``(batch, lead, siglen)``.
        fs : int
            Sampling frequency of the ECG signal.

        Returns
        -------
        filtered_sig : :class:`numpy.ndarray`
            Bandpass filtered ECG signal.
        fs : :class:`int`
            Sampling frequency of the filtered ECG signal.

        """
        self._check_sig(sig)
        filtered_sig = preprocess_multi_lead_signal(
            raw_sig=sig,
            fs=fs,
            band_fs=[self.lowcut, self.highcut],
            filter_type=self.filter_type,
            filter_order=self.filter_order,
        )
        return filtered_sig, fs

    def extra_repr_keys(self) -> List[str]:
        return [
            "lowcut",
            "highcut",
            "filter_type",
            "filter_order",
        ] + super().extra_repr_keys()



"""Resample the signal into fixed sampling frequency or length."""

from numbers import Real
from typing import Any, List, Optional, Tuple

import numpy as np
import scipy.signal as SS

from .base import PreProcessor


__all__ = [
    "Resample",
]


class Resample(PreProcessor):
    """Resample the signal into fixed sampling frequency or length.

    Parameters
    ----------
    fs : int, optional
        Sampling frequency of the resampled ECG.
    siglen : int, optional
        Number of samples in the resampled ECG.

    NOTE
    ----
    One and only one of `fs` and `siglen` should be set.

    Examples
    --------
    .. code-block:: python

        from torch_ecg.cfg import DEFAULTS
        sig = DEFAULTS.RNG.randn(1000)
        pp = Resample(fs=500)
        sig, _ = pp(sig, 250)

    """

    __name__ = "Resample"

    def __init__(
        self, fs: Optional[int] = None, siglen: Optional[int] = None, **kwargs: Any
    ) -> None:
        self.fs = fs
        self.siglen = siglen
        assert (
            sum([bool(self.fs), bool(self.siglen)]) == 1
        ), "one and only one of `fs` and `siglen` should be set"

    def apply(self, sig: np.ndarray, fs: Real) -> Tuple[np.ndarray, int]:
        """Apply the preprocessor to `sig`.

        Parameters
        ----------
        sig : numpy.ndarray
            The ECG signal, can be
                - 1d array, which is a single-lead ECG;
                - 2d array, which is a multi-lead ECG of "lead_first" format;
                - 3d array, which is a tensor of several ECGs, of shape ``(batch, lead, siglen)``.
        fs : numbers.Real
            Sampling frequency of the ECG signal.

        Returns
        -------
        rsmp_sig : :class:`numpy.ndarray`
            The resampled ECG signal.
        new_fs : :class:`int`,
            Sampling frequency of the resampled ECG signal.

        """
        self._check_sig(sig)
        if self.fs is not None:
            rsmp_sig = SS.resample_poly(sig, up=self.fs, down=fs, axis=-1)
            new_fs = self.fs
        else:  # self.siglen is not None
            rsmp_sig = SS.resample(sig, num=self.siglen, axis=-1)
            new_fs = int(round(self.siglen / sig.shape[-1] * fs))
        return rsmp_sig, new_fs

    def extra_repr_keys(self) -> List[str]:
        return [
            "fs",
            "siglen",
        ] + super().extra_repr_keys()



"""
neural architecture search

TODO: replace this module with Ray Tune
"""

from typing import Sequence

import torch
import torch.nn as nn
from torch.nn.parallel import DataParallel as DP
from torch.utils.data.dataset import Dataset

from ..cfg import CFG
from .trainer import BaseTrainer


__all__ = [
    "NAS",
]


class NAS:
    """ """

    __name__ = "NAS"

    def __init__(
        self,
        trainer_cls: BaseTrainer,
        model_cls: nn.Module,
        dataset_cls: Dataset,
        train_config: dict,
        model_configs: Sequence[dict],
        lazy: bool = False,
    ) -> None:
        """
        Parameters
        ----------
        trainer_cls: BaseTrainer,
            trainer class
        model_cls: nn.Module,
            model class
        dataset_cls: Dataset,
            dataset class
        train_config: dict,
            train configurations
        model_configs: sequence of dict,
            model configurations, each with a different network architecture
        lazy: bool, default False,
            whether to load the dataset in the trainer at initialization

        """
        self.trainer_cls = trainer_cls
        self.model_cls = model_cls
        self.dataset_cls = dataset_cls
        self.train_config = CFG(train_config)
        self.model_configs = model_configs
        self.lazy = lazy
        if not lazy:
            self.ds_train = self.dataset_cls(
                self.train_config, training=True, lazy=False
            )
            self.ds_val = self.dataset_cls(
                self.train_config, training=False, lazy=False
            )
        else:
            self.ds_train = None
            self.ds_val = None

    def search(self) -> None:
        """ """
        if self.ds_train is None or self.ds_val is None:
            raise ValueError("training dataset or validation dataset is not set")
        device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        for model_config in self.model_configs:
            model = self.model_cls(
                classes=self.train_config.classes,
                n_leads=self.train_config.n_leads,
                config=model_config,
            )
            if torch.cuda.device_count() > 1:
                model = DP(model)
                # model = DDP(model)
            model.to(device=device)
            model.train()
            trainer = self.trainer_cls(
                model=model,
                dataset_cls=self.dataset_cls,
                train_config=self.train_config,
                model_config=model_config,
                device=device,
                lazy=True,
            )
            trainer._setup_dataloaders(self.ds_train, self.ds_val)
            trainer.train()

            del model
            del trainer
            torch.cuda.empty_cache()

    def _setup_dataset(self, ds_train: Dataset, ds_val: Dataset) -> None:
        """
        Parameters
        ----------
        ds_train: Dataset,
            training dataset
        ds_val: Dataset,
            validation dataset

        """
        self.ds_train = ds_train
        self.ds_val = ds_val



"""
"""

import json
import warnings
from copy import deepcopy
from random import randint, shuffle
from typing import List, Optional, Sequence, Tuple, Any

import numpy as np
from torch.utils.data.dataset import Dataset
from tqdm.auto import tqdm

from ...._preprocessors import PreprocManager
from ....cfg import CFG
from ....databases import LUDB as LR
from ....utils.misc import ReprMixin


__all__ = [
    "LUDBDataset",
]


class LUDBDataset(ReprMixin, Dataset):
    """Data generator for feeding data into pytorch models
    using the :class:`~torch_ecg.databases.LUDB` database.

    Parameters
    ----------
    config : dict
        Configurations for the dataset, ref. `LUDBTrainCfg`.
    training : bool, default True
        If True, the training set will be loaded,
        otherwise the test (val) set will be loaded.
    lazy : bool, default True
        If True, the data will not be loaded immediately,
        instead, it will be loaded on demand.
    **reader_kwargs : dict, optional
        Keyword arguments for the database reader class.

    """

    __name__ = "LUDBDataset"

    def __init__(
        self,
        config: CFG,
        training: bool = True,
        lazy: bool = False,
        **reader_kwargs: Any,
    ) -> None:
        super().__init__()
        self.config = deepcopy(config)
        if reader_kwargs.pop("db_dir", None) is not None:
            warnings.warn(
                "`db_dir` is specified in both config and reader_kwargs", RuntimeWarning
            )
        self.reader = LR(db_dir=self.config.db_dir, **reader_kwargs)
        self.config.db_dir = self.reader.db_dir
        self.training = training
        self.dtype = self.config.np_dtype
        self.classes = self.config.classes
        self.n_classes = len(self.classes)
        self.siglen = self.config.input_len
        if self.config.leads is None:
            self.leads = self.reader.all_leads
        elif isinstance(self.config.leads, str):
            self.leads = [self.config.leads]
        else:
            self.leads = list(self.config.leads)
        self.lazy = lazy

        self.ppm = PreprocManager.from_config(self.config)
        self.records = self._train_test_split(self.config.train_ratio)
        self.fdr = _FastDataReader(self.reader, self.records, self.config)
        self.waveform_priority = ["N", "t", "p", "i"]

        self._signals = None
        self._labels = None
        if not self.lazy:
            self._load_all_data()

    def __len__(self) -> int:
        if self.config.use_single_lead:
            return len(self.leads) * len(self.records)
        return len(self.records)

    def __getitem__(self, index: int) -> Tuple[np.ndarray, np.ndarray]:
        if self.config.use_single_lead:
            rec_idx, lead_idx = divmod(index, len(self.leads))
        else:
            rec_idx, lead_idx = index, None
        rec = self.records[rec_idx]
        if not self.lazy:
            signals = self._signals[rec_idx]
            labels = self._labels[rec_idx]
        else:
            signals, labels = self.fdr[rec_idx]
        if lead_idx is not None:
            signals = signals[[lead_idx], ...]
            labels = labels[lead_idx, ...]
        else:
            # merge labels in all leads to one
            # TODO: map via self.waveform_priority
            labels = np.max(labels, axis=0)
        sampfrom = randint(
            self.config.start_from, signals.shape[1] - self.config.end_at - self.siglen
        )
        sampto = sampfrom + self.siglen
        signals = signals[..., sampfrom:sampto]
        labels = labels[sampfrom:sampto, ...]

        return signals, labels

    def _load_all_data(self) -> None:
        """Load all data into memory."""
        self._signals, self._labels = [], []

        with tqdm(
            self.fdr, total=len(self.fdr), dynamic_ncols=True, mininterval=1.0
        ) as bar:
            for signals, labels in bar:
                self._signals.append(signals)
                self._labels.append(labels)

        self._signals = np.array(self._signals)
        self._labels = np.array(self._labels)

    @property
    def signals(self) -> np.ndarray:
        """Cached signals, only available when `lazy=False`
        or preloading is performed manually.
        """
        return self._signals

    @property
    def labels(self) -> np.ndarray:
        """Cached labels, only available when `lazy=False`
        or preloading is performed manually.
        """
        return self._labels

    def _train_test_split(
        self, train_ratio: float = 0.8, force_recompute: bool = False
    ) -> List[str]:
        """Perform train-test split.

        Parameters
        ----------
        train_ratio : float, default 0.8
            ratio of the train set in the whole dataset.
        force_recompute : bool, default False
            If True, the train-test split will be recomputed,
            regardless of the existing ones stored in json files.

        Returns
        -------
        List[str]
            The list of the records split for training or validation.

        """
        _train_ratio = int(train_ratio * 100)
        _test_ratio = 100 - _train_ratio
        assert _train_ratio * _test_ratio > 0
        train_file = self.reader.db_dir / f"train_ratio_{_train_ratio}.json"
        test_file = self.reader.db_dir / f"test_ratio_{_test_ratio}.json"

        if self.reader._subsample is not None:
            force_recompute = True

        if force_recompute or not all([train_file.is_file(), test_file.is_file()]):
            all_records = deepcopy(self.reader.all_records)
            shuffle(all_records)
            split_idx = int(_train_ratio * len(all_records) / 100)
            train_set = all_records[:split_idx]
            test_set = all_records[split_idx:]
            if self.reader._subsample is None:
                train_file.write_text(json.dumps(train_set, ensure_ascii=False))
                test_file.write_text(json.dumps(test_set, ensure_ascii=False))
        else:
            train_set = json.loads(train_file.read_text())
            test_set = json.loads(test_file.read_text())
        if self.training:
            records = train_set
        else:
            records = test_set
        if self.config.over_sampling > 1:
            records = records * self.config.over_sampling
            shuffle(records)
        return records

    def extra_repr_keys(self) -> List[str]:
        return [
            "training",
            "reader",
        ]


class _FastDataReader(ReprMixin, Dataset):
    """Fast data reader.

    Parameters
    ----------
    reader : CR
        The reader to read the data.
    records : Sequence[str]
        The list of records to read.
    config : CFG
        The configuration.
    ppm : PreprocManager, optional
        The preprocessor manager.

    """

    def __init__(
        self,
        reader: LR,
        records: Sequence[str],
        config: CFG,
        ppm: Optional[PreprocManager] = None,
    ) -> None:
        self.reader = reader
        self.records = records
        self.config = config
        self.ppm = ppm
        self.dtype = self.config.np_dtype

        if self.config.leads is None:
            self.leads = self.reader.all_leads
        elif isinstance(self.config.leads, str):
            self.leads = [self.config.leads]
        else:
            self.leads = list(self.config.leads)

    def __len__(self) -> int:
        return len(self.records)

    def __getitem__(self, index: int) -> Tuple[np.ndarray, np.ndarray]:
        rec = self.records[index]
        signals = self.reader.load_data(
            rec,
            data_format="channel_first",
            units="mV",
        ).astype(self.dtype)
        if self.ppm:
            signals, _ = self.ppm(signals, self.config.fs)
        masks = self.reader.load_masks(
            rec,
            leads=self.leads,
            mask_format="channel_first",
            class_map=self.config.class_map,
        ).astype(self.dtype)
        if self.config.loss == "CrossEntropyLoss":
            return signals, masks
        # expand masks to have n vectors, with n = n_classes
        labels = np.ones(
            (*masks.shape, len(self.config.mask_class_map)), dtype=self.dtype
        )
        for i in range(len(self.leads)):
            for key, val in self.config.mask_class_map.items():
                labels[i, ..., val] = (
                    masks[i, ...] == self.config.class_map[key]
                ).astype(self.dtype)
        return signals, labels

    def extra_repr_keys(self) -> List[str]:
        return [
            "reader",
            "ppm",
        ]



# -*- coding: utf-8 -*-

import math
from copy import deepcopy
from pathlib import Path
from typing import Any, Dict, List, Optional, Sequence, Union

import numpy as np
import wfdb

from ...cfg import CFG
from ...utils.misc import add_docstring, get_record_list_recursive
from ...utils.utils_interval import generalized_intervals_intersection
from ..base import DEFAULT_FIG_SIZE_PER_SEC, PhysioNetDataBase, DataBaseInfo


__all__ = [
    "AFDB",
]


_AFDB_INFO = DataBaseInfo(
    title="""
    MIT-BIH Atrial Fibrillation Database
    """,
    about="""
    1. contains 25 long-term (each 10 hours) ECG recordings of human subjects with atrial fibrillation (mostly paroxysmal)
    2. 23 records out of 25 include the two ECG signals, the left 2 records 00735 and 03665 are represented only by the rhythm (.atr) and unaudited beat (.qrs) annotation files
    3. signals are sampled at 250 samples per second with 12-bit resolution over a range of ±10 millivolts, with a typical recording bandwidth of approximately 0.1 Hz to 40 Hz
    4. 4 classes of rhythms are annotated:

        - AFIB:  atrial fibrillation
        - AFL:   atrial flutter
        - J:     AV junctional rhythm
        - N:     all other rhythms

    5. rhythm annotations almost all start with "(N", except for 4 which start with '(AFIB', which are all within 1 second (250 samples)
    6. Webpage of the database on PhysioNet [1]_. Paper describing the database [2]_.
    """,
    note="""
    1. beat annotation files (.qrs files) were prepared using an automated detector and have NOT been corrected manually
    2. for some records, manually corrected beat annotation files (.qrsc files) are available
    3. one should never use wfdb.rdann with arguments `sampfrom`, since one has to know the `aux_note` (with values in ["(N", "(J", "(AFL", "(AFIB"]) before the index at `sampfrom`
    """,
    usage=[
        "Atrial fibrillation (AF) detection",
    ],
    references=[
        "https://physionet.org/content/afdb/",
        "Moody GB, Mark RG. A new method for detecting atrial fibrillation using R-R intervals. Computers in Cardiology. 10:227-230 (1983).",
    ],
    doi=[
        "10.13026/C2MW2D",
    ],
)


@add_docstring(_AFDB_INFO.format_database_docstring(), mode="prepend")
class AFDB(PhysioNetDataBase):
    """
    Parameters
    ----------
    db_dir : str or pathlib.Path, optional
        Storage path of the database.
        If not specified, data will be fetched from Physionet.
    working_dir : str, optional
        Working directory, to store intermediate files and log files.
    verbose : int, default 1
        Level of logging verbosity.
    kwargs : dict, optional
        Auxilliary key word arguments.

    """

    __name__ = "AFDB"

    def __init__(
        self,
        db_dir: Optional[Union[str, Path]] = None,
        working_dir: Optional[str] = None,
        verbose: int = 1,
        **kwargs: Any,
    ) -> None:
        super().__init__(
            db_name="afdb",
            db_dir=db_dir,
            working_dir=working_dir,
            verbose=verbose,
            **kwargs,
        )
        self.fs = 250
        self.data_ext = "dat"
        self.ann_ext = "atr"
        self.auto_beat_ann_ext = "qrs"
        self.manual_beat_ann_ext = "qrsc"

        self.all_leads = [
            "ECG1",
            "ECG2",
        ]

        self.special_records = ["00735", "03665"]
        self.qrsc_records = None
        self._ls_rec()

        self.class_map = CFG(AFIB=1, AFL=2, J=3, N=0)  # an extra isoelectric
        self.palette = kwargs.get("palette", None)
        if self.palette is None:
            self.palette = CFG(
                AFIB="blue",
                AFL="red",
                J="yellow",
                # N="green",
                qrs="green",
            )

    def _ls_rec(self, local: bool = True) -> None:
        """
        Find all records (relative path without file extension),
        and save into `self._all_records` for further use.

        Parameters
        ----------
        local : bool, default True
            If True, read from local storage,
            prior to using :func:`wfdb.get_record_list`.

        """
        super()._ls_rec(local=local)
        self._all_records = [
            rec for rec in self._all_records if rec not in self.special_records
        ]
        self.qrsc_records = get_record_list_recursive(
            self.db_dir, self.manual_beat_ann_ext, relative=False
        )
        self.qrsc_records = [Path(rec).stem for rec in self.qrsc_records]

    def load_ann(
        self,
        rec: Union[str, int],
        sampfrom: Optional[int] = None,
        sampto: Optional[int] = None,
        ann_format: str = "interval",
        keep_original: bool = False,
    ) -> Union[Dict[str, list], np.ndarray]:
        """Load annotations (header) from the .hea files.

        Parameters
        ----------
        rec : str or int
            Record name or index of the record in :attr:`all_records`.
        sampfrom : int, optional
            Start index of the annotations to be loaded.
        sampto: int, optional
            End index of the annotations to be loaded.
        ann_format : {"interval", "mask"}
            Format of returned annotation,
            by default "interval", case insensitive.
        keep_original : bool, default False
            If True, when `ann_format` is "interval",
            intervals (in the form [a,b]) will keep the same with the annotation file,
            otherwise subtract `sampfrom` if specified.

        Returns
        -------
        ann : dict or numpy.ndarray
            The annotations in the format of intervals, or in the format of masks.

        """
        fp = str(self.get_absolute_path(rec))
        wfdb_ann = wfdb.rdann(fp, extension=self.ann_ext)
        header = wfdb.rdheader(fp)
        sig_len = header.sig_len
        sf = sampfrom or 0
        st = sampto or sig_len
        assert st > sf, "`sampto` should be greater than `sampfrom`!"

        ann = CFG({k: [] for k in self.class_map.keys()})
        critical_points = wfdb_ann.sample.tolist() + [sig_len]
        aux_note = wfdb_ann.aux_note
        if aux_note[0] == "(N":
            # ref. the doc string of the class
            critical_points[0] = 0
        else:
            critical_points.insert(0, 0)
            aux_note.insert(0, "(N")

        for idx, rhythm in enumerate(aux_note):
            ann[rhythm.replace("(", "")].append(
                [critical_points[idx], critical_points[idx + 1]]
            )
        ann = CFG(
            {
                k: generalized_intervals_intersection(l_itv, [[sf, st]])
                for k, l_itv in ann.items()
            }
        )

        if ann_format.lower() == "mask":
            tmp = deepcopy(ann)
            ann = np.full(shape=(st - sf,), fill_value=self.class_map.N, dtype=int)
            for rhythm, l_itv in tmp.items():
                for itv in l_itv:
                    ann[itv[0] - sf : itv[1] - sf] = self.class_map[rhythm]
        elif not keep_original:
            for k, l_itv in ann.items():
                ann[k] = [[itv[0] - sf, itv[1] - sf] for itv in l_itv]

        return ann

    def load_beat_ann(
        self,
        rec: Union[str, int],
        sampfrom: Optional[int] = None,
        sampto: Optional[int] = None,
        use_manual: bool = True,
        keep_original: bool = False,
    ) -> np.ndarray:
        """Load beat annotations from corresponding annotation files.

        Parameters
        ----------
        rec : str or int
            Record name or index of the record in :attr:`all_records`.
        sampfrom : int, optional
            Start index of the annotations to be loaded.
        sampto : int, optional
            End index of the annotations to be loaded.
        use_manual : bool, default True
            If True, use manually annotated beat annotations (qrs),
            instead of those generated by algorithms.
        keep_original : bool, default False
            If True, indices will keep the same with the annotation file,
            otherwise subtract `sampfrom` if specified.

        Returns
        -------
        ann : numpy.ndarray
            Locations (indices) of the qrs complexes.

        """
        if isinstance(rec, int):
            rec = self[rec]
        fp = str(self.get_absolute_path(rec))
        if use_manual and rec in self.qrsc_records:
            ext = self.manual_beat_ann_ext
        else:
            ext = self.auto_beat_ann_ext
        ann = wfdb.rdann(
            fp,
            extension=ext,
            sampfrom=sampfrom or 0,
            sampto=sampto,
        )
        ann = ann.sample
        if not keep_original and sampfrom is not None:
            ann -= sampfrom
        return ann

    @add_docstring(load_beat_ann.__doc__)
    def load_rpeak_indices(
        self,
        rec: Union[str, int],
        sampfrom: Optional[int] = None,
        sampto: Optional[int] = None,
        use_manual: bool = True,
        keep_original: bool = False,
    ) -> np.ndarray:
        """
        alias of `self.load_beat_ann`
        """
        return self.load_beat_ann(rec, sampfrom, sampto, use_manual, keep_original)

    def plot(
        self,
        rec: Union[str, int],
        data: Optional[np.ndarray] = None,
        ann: Optional[Dict[str, np.ndarray]] = None,
        rpeak_inds: Optional[Union[Sequence[int], np.ndarray]] = None,
        ticks_granularity: int = 0,
        leads: Optional[Union[str, int, List[str], List[int]]] = None,
        sampfrom: Optional[int] = None,
        sampto: Optional[int] = None,
        same_range: bool = False,
        **kwargs: Any,
    ) -> None:
        """
        Plot the signals of a record or external signals (units in μV),
        with metadata (fs, labels, tranche, etc.),
        possibly also along with wave delineations.

        Parameters
        ----------
        rec : str or int
            Record name or index of the record in :attr:`all_records`.
        data : numpy.ndarray, optional
            (2-lead) ECG signal to plot,
            should be of the format "channel_first", and compatible with `leads`.
            If given, data of `rec` will not be used,
            which is useful when plotting filtered data.
        ann : dict, optional
            Annotations for `data`, covering those from annotation files,
            in the form of ``{"AFIB":l_itv, "AFL":l_itv, "J":l_itv, "N":l_itv}``,
            where ``l_itv`` in the form of ``[[a, b], ...]``.
            Ignored if `data` is None.
        rpeak_inds : array_like, optional
            Indices of R peaks, covering those from annotation files.
            If `data` is None, then indices should be the absolute indices in the record.
        ticks_granularity : int, default 0
            Granularity to plot axis ticks, the higher the more ticks.
            0 (no ticks) --> 1 (major ticks) --> 2 (major + minor ticks).
        leads : str or int or List[str] or List[int], optional
            The leads to plot.
        sampfrom : int, optional
            Start index of the data to plot.
        sampto : int, optional
            End index of the data to plot.
        same_range : bool, default False
            If True, forces all leads to have the same y range.
        kwargs : dict, optional
            Keyword arguments for :func:`matplotlib.pyplot.plot`, etc.

        """
        if isinstance(rec, int):
            rec = self[rec]
        if "plt" not in dir():
            import matplotlib.pyplot as plt

            plt.MultipleLocator.MAXTICKS = 3000
        if leads is None or leads == "all":
            _leads = self.all_leads
        elif isinstance(leads, str):
            _leads = [leads]
        elif isinstance(leads, int):
            _leads = [self.all_leads[leads]]
        else:
            _leads = leads
        assert all([ld in self.all_leads for ld in _leads]) or set(_leads) <= {0, 1}

        try:
            lead_indices = [self.all_leads.index(ld) for ld in _leads]
        except ValueError:
            lead_indices = _leads
        if data is None:
            _data = self.load_data(
                rec,
                leads=_leads,
                sampfrom=sampfrom,
                sampto=sampto,
                data_format="channel_first",
                units="μV",
            )
        else:
            units = self._auto_infer_units(data)
            self.logger.info(f"input data is auto detected to have units in `{units}`")
            if units.lower() == "mv":
                _data = 1000 * data
            else:
                _data = data
            _leads = [f"ECG_{idx}" for idx in range(_data.shape[0])]
        if ann is None and data is None:
            _ann = self.load_ann(
                rec,
                sampfrom=sampfrom,
                sampto=sampto,
                ann_format="interval",
                keep_original=False,
            )
        else:
            _ann = ann or CFG({k: [] for k in self.class_map.keys()})
        # indices to time
        _ann = {
            k: [[itv[0] / self.fs, itv[1] / self.fs] for itv in l_itv]
            for k, l_itv in _ann.items()
        }
        if rpeak_inds is None and data is None:
            _rpeak = self.load_rpeak_indices(
                rec,
                sampfrom=sampfrom,
                sampto=sampto,
                use_manual=True,
                keep_original=False,
            )
            _rpeak = _rpeak / self.fs  # indices to time
        else:
            _rpeak = np.array(rpeak_inds or []) / self.fs  # indices to time

        ann_plot_alpha = 0.2
        rpeaks_plot_alpha = 0.8

        nb_leads = len(_leads)

        line_len = self.fs * 25  # 25 seconds
        nb_lines = math.ceil(_data.shape[1] / line_len)

        for seg_idx in range(nb_lines):
            seg_data = _data[..., seg_idx * line_len : (seg_idx + 1) * line_len]
            secs = (np.arange(seg_data.shape[1]) + seg_idx * line_len) / self.fs
            seg_ann = {
                k: generalized_intervals_intersection(l_itv, [[secs[0], secs[-1]]])
                for k, l_itv in _ann.items()
            }
            seg_rpeaks = _rpeak[np.where((_rpeak >= secs[0]) & (_rpeak < secs[-1]))[0]]
            fig_sz_w = int(
                round(DEFAULT_FIG_SIZE_PER_SEC * seg_data.shape[1] / self.fs)
            )
            if same_range:
                y_ranges = (
                    np.ones((seg_data.shape[0],)) * np.max(np.abs(seg_data)) + 100
                )
            else:
                y_ranges = np.max(np.abs(seg_data), axis=1) + 100
            fig_sz_h = 6 * y_ranges / 1500
            fig, axes = plt.subplots(
                nb_leads, 1, sharex=True, figsize=(fig_sz_w, np.sum(fig_sz_h))
            )
            if nb_leads == 1:
                axes = [axes]
            for idx in range(nb_leads):
                axes[idx].plot(
                    secs, seg_data[idx], color="black", label=f"lead - {_leads[idx]}"
                )
                axes[idx].axhline(y=0, linestyle="-", linewidth="1.0", color="red")
                # NOTE that `Locator` has default `MAXTICKS` equal to 1000
                if ticks_granularity >= 1:
                    axes[idx].xaxis.set_major_locator(plt.MultipleLocator(0.2))
                    axes[idx].yaxis.set_major_locator(plt.MultipleLocator(500))
                    axes[idx].grid(
                        which="major", linestyle="-", linewidth="0.5", color="red"
                    )
                if ticks_granularity >= 2:
                    axes[idx].xaxis.set_minor_locator(plt.MultipleLocator(0.04))
                    axes[idx].yaxis.set_minor_locator(plt.MultipleLocator(100))
                    axes[idx].grid(
                        which="minor", linestyle=":", linewidth="0.5", color="black"
                    )
                for k, l_itv in seg_ann.items():
                    if k == "N":
                        continue
                    for itv in l_itv:
                        axes[idx].axvspan(
                            itv[0],
                            itv[1],
                            color=self.palette[k],
                            alpha=ann_plot_alpha,
                            label=k,
                        )
                for ri in seg_rpeaks:
                    axes[idx].axvspan(
                        ri - 0.01,
                        ri + 0.01,
                        color=self.palette["qrs"],
                        alpha=rpeaks_plot_alpha,
                    )
                    axes[idx].axvspan(
                        ri - 0.075,
                        ri + 0.075,
                        color=self.palette["qrs"],
                        alpha=ann_plot_alpha,
                    )
                axes[idx].legend(loc="upper left")
                axes[idx].set_xlim(secs[0], secs[-1])
                axes[idx].set_ylim(-y_ranges[idx], y_ranges[idx])
                axes[idx].set_xlabel("Time [s]")
                axes[idx].set_ylabel("Voltage [μV]")
            plt.subplots_adjust(hspace=0.2)
            plt.show()

    @property
    def database_info(self) -> DataBaseInfo:
        return _AFDB_INFO



"""Manger for the augmenters"""

import warnings
from random import sample
from typing import Any, List, Optional, Sequence, Tuple, Union

import torch
from torch import Tensor

from ..utils.misc import default_class_repr, add_docstring
from .base import Augmenter, _augmenter_forward_doc
from .baseline_wander import BaselineWanderAugmenter
from .label_smooth import LabelSmooth
from .mixup import Mixup
from .random_flip import RandomFlip
from .random_masking import RandomMasking
from .random_renormalize import RandomRenormalize
from .stretch_compress import StretchCompress


__all__ = [
    "AugmenterManager",
]


class AugmenterManager(torch.nn.Module):
    """The :class:`~torch.nn.Module` to manage the augmenters.

    Parameters
    ----------
    aug : Tuple[Augmenter], optional
        The augmenters to be added to the manager.
    random : bool, default False
        Whether to apply the augmenters in random order.

    Examples
    --------
    .. code-block:: python

        import torch
        from torch_ecg.cfg import CFG
        from torch_ecg.augmenters import AugmenterManager

        config = CFG(
            random=False,
            fs=500,
            baseline_wander={},
            label_smooth={},
            mixup={},
            random_flip={},
            random_masking={},
            random_renormalize={},
            stretch_compress={},
        )
        am = AugmenterManager.from_config(config)
        sig = torch.randn(32, 12, 5000)
        label = torch.randint(0, 2, (32, 26), dtype=torch.float32)
        mask1 = torch.randint(0, 2, (32, 5000, 3), dtype=torch.float32)
        mask2 = torch.randint(0, 3, (32, 5000), dtype=torch.long)
        sig, label, mask1, mask2 = am(sig, label, mask1, mask2)

    """

    __name__ = "AugmenterManager"

    def __init__(
        self, *augs: Optional[Tuple[Augmenter, ...]], random: bool = False
    ) -> None:
        super().__init__()
        self.random = random
        self._augmenters = list(augs)

    def _add_baseline_wander(self, **config: dict) -> None:
        """Add the baseline wander augmenter to the manager.

        Parameters
        ----------
        **config : dict
            The configuration for the baseline wander augmenter.

        """
        self._augmenters.append(BaselineWanderAugmenter(**config))

    def _add_label_smooth(self, **config: dict) -> None:
        """Add the label smooth augmenter to the manager.

        Parameters
        ----------
        **config : dict
            The configuration for the label smooth augmenter.

        """
        self._augmenters.append(LabelSmooth(**config))

    def _add_mixup(self, **config: dict) -> None:
        """Add the mixup augmenter to the manager.

        Parameters
        ----------
        **config : dict
            The configuration for the mixup augmenter.

        """
        self._augmenters.append(Mixup(**config))

    def _add_random_flip(self, **config: dict) -> None:
        """Add the random flip augmenter to the manager.

        Parameters
        ----------
        **config : dict
            The configuration for the random flip augmenter.

        """
        self._augmenters.append(RandomFlip(**config))

    def _add_random_masking(self, **config: dict) -> None:
        """Add the random masking augmenter to the manager.

        Parameters
        ----------
        **config : dict
            The configuration for the random masking augmenter.

        """
        self._augmenters.append(RandomMasking(**config))

    def _add_random_renormalize(self, **config: dict) -> None:
        """Add the random renormalize augmenter to the manager.

        Parameters
        ----------
        **config : dict
            The configuration for the random renormalize augmenter.

        """
        self._augmenters.append(RandomRenormalize(**config))

    def _add_stretch_compress(self, **config: dict) -> None:
        """Add the stretch compress augmenter to the manager.

        Parameters
        ----------
        **config : dict
            The configuration for the stretch compress augmenter.

        """
        self._augmenters.append(StretchCompress(**config))

    @add_docstring(
        _augmenter_forward_doc.replace(
            "Forward method of the augmenter.",
            "Forward the input ECGs through the augmenters.",
        )
    )
    def forward(
        self,
        sig: Tensor,
        label: Optional[Tensor],
        *extra_tensors: Sequence[Tensor],
        **kwargs: Any,
    ) -> Union[Tensor, Tuple[Tensor]]:
        if len(self.augmenters) == 0:
            # raise ValueError("No augmenters added to the manager.")
            return (sig, label, *extra_tensors)
        ordering = list(range(len(self.augmenters)))
        if self.random:
            ordering = sample(ordering, len(ordering))
        for idx in ordering:
            sig, label, *extra_tensors = self.augmenters[idx](
                sig, label, *extra_tensors, **kwargs
            )
        return (sig, label, *extra_tensors)

    @property
    def augmenters(self) -> List[Augmenter]:
        """The list of augmenters in the manager."""
        return self._augmenters

    def extra_repr(self) -> str:
        """Extra keys for :meth:`__repr__` and :meth:`__str__`."""
        indent = 4 * " "
        s = (
            f"augmenters = [\n{indent}"
            + f",\n{2*indent}".join(
                default_class_repr(aug, depth=2) for aug in self.augmenters
            )
            + f"{indent}\n]"
        )
        return s

    @classmethod
    def from_config(cls, config: dict) -> "AugmenterManager":
        """Create an :class:`AugmenterManager` from a configuration.

        Parameters
        ----------
        config : dict
            The configuration of the augmenters,
            better to be an :class:`~collections.OrderedDict`.

        Returns
        -------
        am : :class:`AugmenterManager`
            A new instance of :class:`AugmenterManager`.

        """
        am = cls(random=config.get("random", False))
        _mapping = {
            "baseline_wander": am._add_baseline_wander,
            "label_smooth": am._add_label_smooth,
            "mixup": am._add_mixup,
            "random_flip": am._add_random_flip,
            "random_masking": am._add_random_masking,
            "random_renormalize": am._add_random_renormalize,
            "stretch_compress": am._add_stretch_compress,
        }
        for aug_name, aug_config in config.items():
            if aug_name in [
                "fs",
                "random",
            ]:
                continue
            elif aug_name in _mapping and isinstance(aug_config, dict):
                _mapping[aug_name](fs=config["fs"], **aug_config)
            else:
                # just ignore the other items
                pass
                # raise ValueError(f"Unknown augmenter name: {aug_name}")
        return am

    def rearrange(self, new_ordering: List[str]) -> None:
        """Rearrange the augmenters in the manager.

        Parameters
        ----------
        new_ordering : List[str]
            The list of augmenter names in the new order.

        """
        if self.random:
            warnings.warn(
                "The augmenters are applied in random order, "
                "rearranging the augmenters will not take effect.",
                RuntimeWarning,
            )
        _mapping = {  # built-in augmenters
            "".join([w.capitalize() for w in k.split("_")]): k
            for k in "label_smooth,mixup,random_flip,random_masking,random_renormalize,stretch_compress".split(
                ","
            )
        }
        _mapping.update({"BaselineWanderAugmenter": "baseline_wander"})
        _mapping.update({v: k for k, v in _mapping.items()})
        for k in new_ordering:
            if k not in _mapping:
                # allow custom augmenters
                assert k in [
                    am.__class__.__name__ for am in self._augmenters
                ], f"Unknown augmenter name: `{k}`"
                _mapping.update({k: k})
        assert len(new_ordering) == len(set(new_ordering)), "Duplicate augmenter names."
        assert len(new_ordering) == len(
            self._augmenters
        ), "Number of augmenters mismatch."

        self._augmenters.sort(
            key=lambda aug: new_ordering.index(_mapping[aug.__class__.__name__])
        )



"""
Use median filter to remove baseline.
Note that highpass filters also have the effect of baseline removal
"""

import warnings
from numbers import Real
from typing import Any, List, Tuple

import numpy as np

from .base import PreProcessor, preprocess_multi_lead_signal


__all__ = [
    "BaselineRemove",
]


class BaselineRemove(PreProcessor):
    """Baseline removal using median filter.

    Parameters
    ----------
    window1 : float, default 0.2
        The smaller window size of the median filter, with units in seconds.
    highcut : float, default 0.6
        The larger window size of the median filter, with units in seconds.

    Examples
    --------
    .. code-block:: python

        from torch_ecg.cfg import DEFAULTS
        sig = DEFAULTS.RNG.randn(1000)
        pp = BaselineRemove(window1=0.2, window2=0.6)
        sig, _ = pp(sig, 500)

    """

    __name__ = "BaselineRemove"

    def __init__(
        self, window1: float = 0.2, window2: float = 0.6, **kwargs: Any
    ) -> None:
        self.window1 = window1
        self.window2 = window2
        if self.window2 < self.window1:
            self.window1, self.window2 = self.window2, self.window1
            warnings.warn(
                "values of `window1` and `window2` are switched", RuntimeWarning
            )

    def apply(self, sig: np.ndarray, fs: Real) -> Tuple[np.ndarray, int]:
        """Apply the preprocessor to `sig`.

        Parameters
        ----------
        sig : numpy.ndarray
            The ECG signal, can be
                - 1d array, which is a single-lead ECG;
                - 2d array, which is a multi-lead ECG of "lead_first" format;
                - 3d array, which is a tensor of several ECGs, of shape ``(batch, lead, siglen)``.
        fs : numbers.Real
            Sampling frequency of the ECG signal.

        Returns
        -------
        filtered_sig : :class:`numpy.ndarray`
            The median filtered (hence baseline removed) ECG signal.
        fs : :class:`int`
            Sampling frequency of the filtered ECG signal.

        """
        self._check_sig(sig)
        filtered_sig = preprocess_multi_lead_signal(
            raw_sig=sig,
            fs=fs,
            bl_win=[self.window1, self.window2],
        )
        return filtered_sig, fs

    def extra_repr_keys(self) -> List[str]:
        return ["window1", "window2"] + super().extra_repr_keys()
